# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Tuple

import torch
import torch.nn as nn
from mmcv.cnn.bricks import ConvModule
from mmdet.structures.bbox.transforms import (bbox_cxcywh_to_xyxy,
                                              bbox_xyxy_to_cxcywh)
from mmengine.model import BaseModule
from mmengine.structures import InstanceData
from torch import Tensor

from mmtrack.registry import MODELS, TASK_UTILS
from mmtrack.utils import InstanceList, SampleList
from ..task_modules.track import depthwise_correlation


@MODELS.register_module()
class CorrelationHead(BaseModule):
    """Correlation head module.

    This module is proposed in
    "SiamRPN++: Evolution of Siamese Visual Tracking with Very Deep Networks.
    `SiamRPN++ <https://arxiv.org/abs/1812.11703>`_.

    Args:
        in_channels (int): Input channels.
        mid_channels (int): Middle channels.
        out_channels (int): Output channels.
        kernel_size (int): Kernel size of convs. Defaults to 3.
        norm_cfg (dict): Configuration of normlization method after each conv.
            Defaults to dict(type='BN').
        act_cfg (dict): Configuration of activation method after each conv.
            Defaults to dict(type='ReLU').
        init_cfg (Optional(dict)): Initialization config dict.
            Defaults to None.
    """

    def __init__(self,
                 in_channels: int,
                 mid_channels: int,
                 out_channels: int,
                 kernel_size: int = 3,
                 norm_cfg: dict = dict(type='BN'),
                 act_cfg: dict = dict(type='ReLU'),
                 init_cfg: Optional[dict] = None,
                 **kwargs):
        super(CorrelationHead, self).__init__(init_cfg)
        self.kernel_convs = ConvModule(
            in_channels=in_channels,
            out_channels=mid_channels,
            kernel_size=kernel_size,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

        self.search_convs = ConvModule(
            in_channels=in_channels,
            out_channels=mid_channels,
            kernel_size=kernel_size,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

        self.head_convs = nn.Sequential(
            ConvModule(
                in_channels=mid_channels,
                out_channels=mid_channels,
                kernel_size=1,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg),
            ConvModule(
                in_channels=mid_channels,
                out_channels=out_channels,
                kernel_size=1,
                act_cfg=None))

    def forward(self, kernel: Tensor, search: Tensor) -> Tensor:
        """Forward function.

        Args:
            kernel (Tensor): The feature map of the template.
            search (Tensor): The feature map of search images.

        Returns:
            Tensor: The correlation results.
        """
        kernel = self.kernel_convs(kernel)
        search = self.search_convs(search)
        correlation_maps = depthwise_correlation(search, kernel)
        out = self.head_convs(correlation_maps)
        return out


@MODELS.register_module()
class SiameseRPNHead(BaseModule):
    """Siamese RPN head.

    This module is proposed in
    "SiamRPN++: Evolution of Siamese Visual Tracking with Very Deep Networks.
    `SiamRPN++ <https://arxiv.org/abs/1812.11703>`_.

    Args:
        anchor_generator (dict): Configuration to build anchor generator
            module.

        in_channels (int): Input channels.

        kernel_size (int): Kernel size of convs. Defaults to 3.

        norm_cfg (dict): Configuration of normlization method after each conv.
            Defaults to dict(type='BN').

        weighted_sum (bool): If True, use learnable weights to weightedly sum
            the output of multi heads in siamese rpn , otherwise, use
            averaging. Defaults to False.

        bbox_coder (dict): Configuration to build bbox coder. Defaults to
            dict(type='DeltaXYWHBBoxCoder', target_means=[0., 0., 0., 0.],
            target_stds=[1., 1., 1., 1.]).

        loss_cls (dict): Configuration to build classification loss. Defaults
            to dict( type='CrossEntropyLoss', reduction='sum', loss_weight=1.0)

        loss_bbox (dict): Configuration to build bbox regression loss. Defaults
            to dict( type='L1Loss', reduction='sum', loss_weight=1.2).

        train_cfg (Dict): Training setting. Defaults to None.

        test_cfg (Dict): Testing setting. Defaults to None.

        init_cfg (dict or list[dict], optional): Initialization config dict.
            Defaults to None.
    """

    def __init__(self,
                 anchor_generator: dict,
                 in_channels: int,
                 kernel_size: int = 3,
                 norm_cfg: dict = dict(type='BN'),
                 weighted_sum: bool = False,
                 bbox_coder: dict = dict(
                     type='DeltaXYWHBBoxCoder',
                     target_means=[0., 0., 0., 0.],
                     target_stds=[1., 1., 1., 1.]),
                 loss_cls: dict = dict(
                     type='CrossEntropyLoss', reduction='sum',
                     loss_weight=1.0),
                 loss_bbox: dict = dict(
                     type='L1Loss', reduction='sum', loss_weight=1.2),
                 train_cfg: Optional[dict] = None,
                 test_cfg: Optional[dict] = None,
                 init_cfg: Optional[dict] = None,
                 *args,
                 **kwargs):
        super(SiameseRPNHead, self).__init__(init_cfg)
        self.anchor_generator = TASK_UTILS.build(anchor_generator)
        self.bbox_coder = TASK_UTILS.build(bbox_coder)
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        self.assigner = TASK_UTILS.build(self.train_cfg.assigner)
        self.sampler = TASK_UTILS.build(self.train_cfg.sampler)
        self.fp16_enabled = False

        self.cls_heads = nn.ModuleList()
        self.reg_heads = nn.ModuleList()
        for i in range(len(in_channels)):
            self.cls_heads.append(
                CorrelationHead(in_channels[i], in_channels[i],
                                2 * self.anchor_generator.num_base_anchors[0],
                                kernel_size, norm_cfg))
            self.reg_heads.append(
                CorrelationHead(in_channels[i], in_channels[i],
                                4 * self.anchor_generator.num_base_anchors[0],
                                kernel_size, norm_cfg))

        self.weighted_sum = weighted_sum
        if self.weighted_sum:
            self.cls_weight = nn.Parameter(torch.ones(len(in_channels)))
            self.reg_weight = nn.Parameter(torch.ones(len(in_channels)))

        self.loss_cls = MODELS.build(loss_cls)
        self.loss_bbox = MODELS.build(loss_bbox)

    def forward(self, template_feats: Tuple[Tensor, ...],
                search_feats: Tuple[Tensor, ...]) -> Tuple[Tensor, Tensor]:
        """Forward with features `template_feats` of template images and
        features `search_feats` of search images.

        Args:
            template_feats (tuple[Tensor, ...]): Tuple of Tensor with shape
                (N, C, H, W) denoting the multi level feature maps of template
                images. Typically H and W equal to 7.
            search_feats (tuple[Tensor, ...]): Tuple of Tensor with shape
                (N, C, H, W) denoting the multi level feature maps of search
                images. Typically H and W equal to 31.

        Returns:
            tuple(Tensor, Tensor): It contains
              - ``cls_score``: a Tensor with shape
                (N, 2 * num_base_anchors, H, W)
              - ``bbox_pred``: a Tensor with shape
                (N, 4 * num_base_anchors, H, W).
                Typically H and W equal to 25.
        """
        assert isinstance(template_feats, tuple) and isinstance(
            search_feats, tuple)
        assert len(template_feats) == len(search_feats) and len(
            template_feats) == len(self.cls_heads)

        if self.weighted_sum:
            cls_weight = nn.functional.softmax(self.cls_weight, dim=0)
            reg_weight = nn.functional.softmax(self.reg_weight, dim=0)
        else:
            reg_weight = cls_weight = [
                1.0 / len(template_feats) for i in range(len(template_feats))
            ]

        cls_score = 0
        bbox_pred = 0
        for i in range(len(template_feats)):
            cls_score_single = self.cls_heads[i](template_feats[i],
                                                 search_feats[i])
            bbox_pred_single = self.reg_heads[i](template_feats[i],
                                                 search_feats[i])
            cls_score += cls_weight[i] * cls_score_single
            bbox_pred += reg_weight[i] * bbox_pred_single

        return cls_score, bbox_pred

    def _get_init_targets(self, bboxes: Tensor,
                          score_maps_size: torch.Size) -> Tuple[Tensor, ...]:
        """Initialize the training targets based on flattened anchors of the
        last score map.

        Args:
            bboxes (Tensor): The generated anchors.
            score_maps_size (torch.Size): denoting the output size
                (height, width) of the network.

        Returns:
            tuple(Tensor, ...): It contains
              - ``labels``: in shape (N, H * W * num_base_anchors)
              - ``labels_weights``: in shape (N, H * W * num_base_anchors)
              - ``bbox_targets``: in shape (N, H * W * num_base_anchors, 4)
              - ``bbox_weights``: in shape (N, H * W * num_base_anchors, 4)
        """
        num_base_anchors = self.anchor_generator.num_base_anchors[0]
        H, W = score_maps_size
        num_anchors = H * W * num_base_anchors
        labels = bboxes.new_zeros((num_anchors, ), dtype=torch.long)
        labels_weights = bboxes.new_zeros((num_anchors, ))
        bbox_weights = bboxes.new_zeros((num_anchors, 4))
        bbox_targets = bboxes.new_zeros((num_anchors, 4))
        return labels, labels_weights, bbox_targets, bbox_weights

    def _get_positive_pair_targets(
            self, gt_instances: InstanceData,
            score_maps_size: torch.Size) -> Tuple[Tensor, ...]:
        """Generate the training targets for positive exemplar image and search
        image pair.

        Args:
            gt_instances (:obj:`InstanceData`): Groundtruth instances. It
                usually includes ``bboxes`` and ``labels`` attributes.
                ``bboxes`` of each search image is of
                shape (1, 4) in [tl_x, tl_y, br_x, br_y] format.
            score_maps_size (torch.Size): denoting the output size
                (height, width) of the network.

        Returns:
            tuple(Tensor, ...): It contains
              - ``labels``: in shape (N, H * W * num_base_anchors)
              - ``labels_weights``: in shape (N, H * W * num_base_anchors)
              - ``bbox_targets``: in shape (N, H * W * num_base_anchors, 4)
              - ``bbox_weights``: in shape (N, H * W * num_base_anchors, 4)
        """
        gt_bbox = gt_instances.bboxes
        (labels, labels_weights, bbox_targets,
         bbox_weights) = self._get_init_targets(gt_bbox, score_maps_size)

        if not hasattr(self, 'anchors'):
            self.anchors = self.anchor_generator.grid_priors(
                [score_maps_size], device=gt_bbox.device)[0]
            # Transform the coordinate origin from the top left corner to the
            # center in the scaled score map.
            feat_h, feat_w = score_maps_size
            stride_w, stride_h = self.anchor_generator.strides[0]
            self.anchors[:, 0:4:2] -= (feat_w // 2) * stride_w
            self.anchors[:, 1:4:2] -= (feat_h // 2) * stride_h

        anchors = self.anchors.clone()

        # The scaled feature map and the searched image have the same center.
        # Transform coordinate origin from the center to the top left corner in
        # the searched image.
        anchors += self.train_cfg.search_size // 2

        pred_instances = InstanceData(priors=anchors)
        assign_result = self.assigner.assign(pred_instances, gt_instances)
        sampling_result = self.sampler.sample(assign_result, pred_instances,
                                              gt_instances)
        pos_inds = sampling_result.pos_inds
        neg_inds = sampling_result.neg_inds
        neg_upper_bound = int(self.sampler.num *
                              (1 - self.sampler.pos_fraction))
        if len(neg_inds) > neg_upper_bound:
            neg_inds = neg_inds[:neg_upper_bound]

        if len(pos_inds) > 0:
            labels[pos_inds] = 1
            labels_weights[pos_inds] = 1.0 / len(pos_inds) / 2
            bbox_weights[pos_inds] = 1.0 / len(pos_inds)

        if len(neg_inds) > 0:
            labels[neg_inds] = 0
            labels_weights[neg_inds] = 1.0 / len(neg_inds) / 2

        bbox_targets[pos_inds, :] = self.bbox_coder.encode(
            sampling_result.pos_priors, sampling_result.pos_gt_bboxes)

        return labels, labels_weights, bbox_targets, bbox_weights

    def _get_negative_pair_targets(
            self, gt_instances: InstanceData,
            score_maps_size: torch.Size) -> Tuple[Tensor, ...]:
        """Generate the training targets for negative exemplar image and search
        image pair.

        Args:
            gt_instances (:obj:`InstanceData`): Groundtruth instances. It
                usually includes ``bboxes`` and ``labels`` attributes.
                ``bboxes`` of each search image is of
                shape (1, 4) in [tl_x, tl_y, br_x, br_y] format.
            score_maps_size (torch.Size): denoting the output size
                (height, width) of the network.

        Returns:
            tuple(Tensor, ...): It contains
              - ``labels``: in shape (N, H * W * num_base_anchors)
              - ``labels_weights``: in shape (N, H * W * num_base_anchors)
              - ``bbox_targets``: in shape (N, H * W * num_base_anchors, 4)
              - ``bbox_weights``: in shape (N, H * W * num_base_anchors, 4)
        """
        gt_bbox = gt_instances.bboxes
        (labels, labels_weights, bbox_targets,
         bbox_weights) = self._get_init_targets(gt_bbox, score_maps_size)
        H, W = score_maps_size
        target_cx, target_cy, _, _ = bbox_xyxy_to_cxcywh(gt_bbox)[0]
        anchor_stride = self.anchor_generator.strides[0]

        cx = W // 2
        cy = H // 2
        cx += int(
            torch.ceil((target_cx - self.train_cfg.search_size // 2) /
                       anchor_stride[0] + 0.5))
        cy += int(
            torch.ceil((target_cy - self.train_cfg.search_size // 2) /
                       anchor_stride[1] + 0.5))

        left = max(0, cx - 3)
        right = min(W, cx + 4)
        top = max(0, cy - 3)
        down = min(H, cy + 4)

        labels = labels.view(H, W, -1)
        labels[...] = -1
        labels[top:down, left:right, :] = 0

        labels = labels.view(-1)
        neg_inds = torch.nonzero(labels == 0, as_tuple=False)[:, 0]
        index = torch.randperm(
            neg_inds.numel(), device=neg_inds.device)[:self.train_cfg.num_neg]
        neg_inds = neg_inds[index]

        labels[...] = -1
        if len(neg_inds) > 0:
            labels[neg_inds] = 0
            labels_weights[neg_inds] = 1.0 / len(neg_inds) / 2

        # TODO: check it whether it's right.
        labels[...] = 0

        return labels, labels_weights, bbox_targets, bbox_weights

    def get_targets(self, batch_gt_instances: InstanceList,
                    score_maps_size: torch.Size) -> Tuple[Tensor, ...]:
        """Generate the training targets for exemplar image and search image
        pairs.

        Args:
            batch_gt_instances (list[InstanceData]): Batch of
                groundtruth instances. It usually includes ``bboxes`` and
                ``labels`` attributes. ``bboxes`` of each search image is of
                shape (1, 4) in [tl_x, tl_y, br_x, br_y] format.
            score_maps_size (torch.Size): denoting the output size
                (height, width) of the network.

        Returns:
            tuple(Tensor, ...): It contains
              - ``all_labels``: in shape (N, H * W * num_base_anchors)
              - ``all_labels_weights``: in shape (N, H * W * num_base_anchors)
              - ``all_bbox_targets``: in shape (N, H * W * num_base_anchors, 4)
              - ``all_bbox_weights``: in shape (N, H * W * num_base_anchors, 4)
        """
        (all_labels, all_labels_weights, all_bbox_targets,
         all_bbox_weights) = [], [], [], []

        for gt_instances in batch_gt_instances:
            is_positive_pair = gt_instances['labels'][0]
            if is_positive_pair:
                (labels, labels_weights, bbox_targets,
                 bbox_weights) = self._get_positive_pair_targets(
                     gt_instances, score_maps_size)
            else:
                (labels, labels_weights, bbox_targets,
                 bbox_weights) = self._get_negative_pair_targets(
                     gt_instances, score_maps_size)

            all_labels.append(labels)
            all_labels_weights.append(labels_weights)
            all_bbox_targets.append(bbox_targets)
            all_bbox_weights.append(bbox_weights)

        all_labels = torch.stack(all_labels)
        all_labels_weights = torch.stack(all_labels_weights) / len(
            all_labels_weights)
        all_bbox_targets = torch.stack(all_bbox_targets)
        all_bbox_weights = torch.stack(all_bbox_weights) / len(
            all_bbox_weights)

        return (all_labels, all_labels_weights, all_bbox_targets,
                all_bbox_weights)

    def loss(self, template_feats: Tuple[Tensor], search_feats: Tuple[Tensor],
             data_samples: SampleList, **kwargs) -> dict:
        """Perform forward propagation and loss calculation of the tracking
        head on the features of the upstream network.

        Args:
            template_feats (tuple[Tensor, ...]): Tuple of Tensor with
                shape (N, C, H, W) denoting the multi level feature maps of
                exemplar images. Typically H and W equal to 7.
            search_feats (tuple[Tensor, ...]): Tuple of Tensor with shape
                (N, C, H, W) denoting the multi level feature maps of
                search images. Typically H and W equal to 31.
            data_samples (List[:obj:`TrackDataSample`]): The Data
                Samples. It usually includes information such as `gt_instance`.

        Returns:
            dict: A dictionary of loss components.
        """
        outs = self(template_feats, search_feats)

        batch_gt_instances = []
        batch_img_metas = []
        batch_search_gt_instances = []
        for data_sample in data_samples:
            batch_img_metas.append(data_sample.metainfo)
            batch_gt_instances.append(data_sample.gt_instances)
            batch_search_gt_instances.append(data_sample.search_gt_instances)

        loss_inputs = outs + (batch_gt_instances, batch_search_gt_instances,
                              batch_img_metas)
        losses = self.loss_by_feat(*loss_inputs)
        return losses

    def loss_by_feat(self, cls_score: Tensor, bbox_pred: Tensor,
                     batch_gt_instances: InstanceList,
                     batch_search_gt_instances: InstanceList,
                     batch_img_metas: List[dict]) -> dict:
        """Compute loss.

        Args:
            cls_score (Tensor): of shape (N, 2 * num_base_anchors, H, W).
            bbox_pred (Tensor): of shape (N, 4 * num_base_anchors, H, W).
            batch_gt_instances (InstanceList): the instances in a batch.
            batch_search_gt_instances (InstanceList): the search instances in a
                batch.
            batch_img_metas (List[dict]): the meta information of all images in
                a batch.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        N, _, H, W = cls_score.shape
        (labels, labels_weights, bbox_targets,
         bbox_weights) = self.get_targets(batch_search_gt_instances, (H, W))

        losses = {}

        cls_score = cls_score.view(N, 2, -1, H, W)
        cls_score = cls_score.permute(0, 3, 4, 2, 1).contiguous().view(-1, 2)
        labels = labels.view(-1)
        labels_weights = labels_weights.view(-1)
        losses['loss_rpn_cls'] = self.loss_cls(
            cls_score, labels, weight=labels_weights)

        bbox_pred = bbox_pred.view(N, 4, -1, H, W)
        bbox_pred = bbox_pred.permute(0, 3, 4, 2, 1).contiguous().view(-1, 4)

        bbox_targets = bbox_targets.view(-1, 4)
        bbox_weights = bbox_weights.view(-1, 4)
        losses['loss_rpn_bbox'] = self.loss_bbox(
            bbox_pred, bbox_targets, weight=bbox_weights)

        return losses

    def predict(self, template_feats: Tuple[Tensor],
                search_feats: Tuple[Tensor], data_samples: SampleList,
                prev_bbox: Tensor, scale_factor: Tensor) -> InstanceList:
        """Perform forward propagation of the tracking head and predict
        tracking results on the features of the upstream network.

        Args:
            template_feats (tuple[Tensor, ...]): Tuple of Tensor with
                shape (N, C, H, W) denoting the multi level feature maps of
                exemplar images. Typically H and W equal to 7.
            search_feats (tuple[Tensor, ...]): Tuple of Tensor with shape
                (N, C, H, W) denoting the multi level feature maps of
                search images. Typically H and W equal to 31.

            data_samples (List[:obj:`TrackDataSample`]): The Data
                Samples. It usually includes information such as `gt_instance`.
            prev_bbox (Tensor): of shape (4, ) in [cx, cy, w, h] format.
            scale_factor (Tensor): scale factor.

        Returns:
            List[:obj:`InstanceData`]: Object tracking results of each image
            after the post process. Each item usually contains following keys.
                - scores (Tensor): Classification scores, has a shape (1, )
                - bboxes (Tensor): Has a shape (1, 4),
                  the last dimension 4 arrange as [x1, y1, x2, y2].
        """
        batch_img_metas = [
            data_samples.metainfo for data_samples in data_samples
        ]
        outs = self(template_feats, search_feats)
        predictions = self.predict_by_feat(
            *outs,
            prev_bbox=prev_bbox,
            scale_factor=scale_factor,
            batch_img_metas=batch_img_metas)
        return predictions

    def predict_by_feat(self, cls_score: Tensor, bbox_pred: Tensor,
                        prev_bbox: Tensor, scale_factor: Tensor,
                        batch_img_metas: List[dict]) -> InstanceList:
        """Track `prev_bbox` to current frame based on the output of network.

        Args:
            cls_score (Tensor): of shape (1, 2 * num_base_anchors, H, W).
            bbox_pred (Tensor): of shape (1, 4 * num_base_anchors, H, W).
            prev_bbox (Tensor): of shape (4, ) in [cx, cy, w, h] format.
            scale_factor (Tensor): scale factor.
            batch_img_metas (List[dict]): the meta information of all images in
                a batch.

        Returns:
            List[:obj:`InstanceData`]: Object tracking results of each image
            after the post process. Each item usually contains following keys.
                - scores (Tensor): Classification scores, has a shape (1, )
                - bboxes (Tensor): Has a shape (1, 4),
                  the last dimension 4 arrange as [x1, y1, x2, y2].
        """
        score_maps_size = [(cls_score.shape[2:])]
        if not hasattr(self, 'anchors'):
            self.anchors = self.anchor_generator.grid_priors(
                score_maps_size, device=cls_score.device)[0]
            # Transform the coordinate origin from the top left corner to the
            # center in the scaled feature map.
            feat_h, feat_w = score_maps_size[0]
            stride_w, stride_h = self.anchor_generator.strides[0]
            self.anchors[:, 0:4:2] -= (feat_w // 2) * stride_w
            self.anchors[:, 1:4:2] -= (feat_h // 2) * stride_h

        if not hasattr(self, 'windows'):
            self.windows = self.anchor_generator.gen_2d_hanning_windows(
                score_maps_size, cls_score.device)[0]

        H, W = score_maps_size[0]
        cls_score = cls_score.view(2, -1, H, W)
        cls_score = cls_score.permute(2, 3, 1, 0).contiguous().view(-1, 2)
        cls_score = cls_score.softmax(dim=1)[:, 1]

        bbox_pred = bbox_pred.view(4, -1, H, W)
        bbox_pred = bbox_pred.permute(2, 3, 1, 0).contiguous().view(-1, 4)
        bbox_pred = self.bbox_coder.decode(self.anchors, bbox_pred)

        result = InstanceData()
        result.scores = cls_score
        result.bboxes = bbox_pred

        return self._bbox_post_process([result],
                                       prev_bbox,
                                       scale_factor,
                                       batch_img_metas=batch_img_metas)

    def _bbox_post_process(self, results: InstanceList, prev_bbox: Tensor,
                           scale_factor: Tensor, batch_img_metas: List[dict],
                           **kwargs) -> InstanceList:
        """The postprocess of tracking bboxes.

        Args:
            results (InstanceList): tracking results.
            prev_bbox (Tensor): of shape (4, ) in [cx, cy, w, h] format.
            scale_factor (Tensor): scale factor.
            batch_img_metas (List[dict]): the meta information of all images in
                a batch.

        Returns:
            List[:obj:`InstanceData`]: Object tracking results of each image
            after the post process. Each item usually contains following keys.
                - scores (Tensor): Classification scores, has a shape (1, )
                - bboxes (Tensor): Has a shape (1, 4),
                  the last dimension 4 arrange as [x1, y1, x2, y2].
        """
        result = results[0]
        bbox_pred = bbox_xyxy_to_cxcywh(result.bboxes)

        def change_ratio(ratio):
            return torch.max(ratio, 1. / ratio)

        def enlarge_size(w, h):
            pad = (w + h) * 0.5
            return torch.sqrt((w + pad) * (h + pad))

        # scale penalty
        scale_penalty = change_ratio(
            enlarge_size(bbox_pred[:, 2], bbox_pred[:, 3]) / enlarge_size(
                prev_bbox[2] * scale_factor, prev_bbox[3] * scale_factor))

        # aspect ratio penalty
        aspect_ratio_penalty = change_ratio(
            (prev_bbox[2] / prev_bbox[3]) /
            (bbox_pred[:, 2] / bbox_pred[:, 3]))

        # penalize cls_score
        penalty = torch.exp(-(aspect_ratio_penalty * scale_penalty - 1) *
                            self.test_cfg.penalty_k)
        penalty_score = penalty * result.scores

        # window penalty
        penalty_score = penalty_score * (1 - self.test_cfg.window_influence) \
            + self.windows * self.test_cfg.window_influence

        best_idx = torch.argmax(penalty_score)
        result = result[best_idx.item()]

        best_bbox = bbox_pred[best_idx, :] / scale_factor
        final_bbox = torch.zeros_like(best_bbox)

        # map the bbox center from the searched image to the original image.
        final_bbox[0] = best_bbox[0] + prev_bbox[0]
        final_bbox[1] = best_bbox[1] + prev_bbox[1]

        # smooth bbox
        lr = penalty[best_idx] * result.scores * self.test_cfg.lr
        final_bbox[2] = prev_bbox[2] * (1 - lr) + best_bbox[2] * lr
        final_bbox[3] = prev_bbox[3] * (1 - lr) + best_bbox[3] * lr

        # clip boundary
        img_shape = batch_img_metas[0]['ori_shape']
        # rather than [x1, x2, y1, y2] format.
        final_bbox = self._bbox_clip(final_bbox, img_shape[0], img_shape[1])

        result.bboxes = bbox_cxcywh_to_xyxy(final_bbox)[None]

        return [result]

    def _bbox_clip(self, bbox: Tensor, img_h: int, img_w: int) -> Tensor:
        """Clip the bbox with [cx, cy, w, h] format.

        Args:
            img (Tensor): of shape (1, C, H, W) encoding original input
                image.
            bbox (Tensor): The given instance bbox of first frame that
                need be tracked in the following frames. The shape of the box
                is of (4, ) shape in [cx, cy, w, h] format.

        Returns:
            Tensor: The clipped boxes.
        """
        bbox[0] = bbox[0].clamp(0., img_w)
        bbox[1] = bbox[1].clamp(0., img_h)
        bbox[2] = bbox[2].clamp(10., img_w)
        bbox[3] = bbox[3].clamp(10., img_h)
        return bbox



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

import torch
from mmdet.structures.bbox import bbox_overlaps
from mmengine.structures import InstanceData
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.structures import TrackDataSample
from .base_tracker import BaseTracker


@MODELS.register_module()
class MaskTrackRCNNTracker(BaseTracker):
    """Tracker for MaskTrack R-CNN.

    Args:
        match_weights (dict[str : float]): The Weighting factor when computing
        the match score. It contains keys as follows:

            - det_score (float): The coefficient of `det_score` when computing
                match score.
            - iou (float): The coefficient of `ious` when computing match
                score.
            - det_label (float): The coefficient of `label_deltas` when
                computing match score.
    """

    def __init__(self,
                 match_weights: dict = dict(
                     det_score=1.0, iou=2.0, det_label=10.0),
                 **kwargs):
        super().__init__(**kwargs)
        self.match_weights = match_weights

    def get_match_score(self, bboxes: Tensor, labels: Tensor, scores: Tensor,
                        prev_bboxes: Tensor, prev_labels: Tensor,
                        similarity_logits: Tensor) -> Tensor:
        """Get the match score.

        Args:
            bboxes (torch.Tensor): of shape (num_current_bboxes, 4) in
                [tl_x, tl_y, br_x, br_y] format. Denoting the detection
                bboxes of current frame.
            labels (torch.Tensor): of shape (num_current_bboxes, )
            scores (torch.Tensor): of shape (num_current_bboxes, )
            prev_bboxes (torch.Tensor): of shape (num_previous_bboxes, 4) in
                [tl_x, tl_y, br_x, br_y] format.  Denoting the detection bboxes
                of previous frame.
            prev_labels (torch.Tensor): of shape (num_previous_bboxes, )
            similarity_logits (torch.Tensor): of shape (num_current_bboxes,
                num_previous_bboxes + 1). Denoting the similarity logits from
                track head.

        Returns:
            torch.Tensor: The matching score of shape (num_current_bboxes,
            num_previous_bboxes + 1)
        """
        similarity_scores = similarity_logits.softmax(dim=1)

        ious = bbox_overlaps(bboxes, prev_bboxes)
        iou_dummy = ious.new_zeros(ious.shape[0], 1)
        ious = torch.cat((iou_dummy, ious), dim=1)

        label_deltas = (labels.view(-1, 1) == prev_labels).float()
        label_deltas_dummy = label_deltas.new_ones(label_deltas.shape[0], 1)
        label_deltas = torch.cat((label_deltas_dummy, label_deltas), dim=1)

        match_score = similarity_scores.log()
        match_score += self.match_weights['det_score'] * \
            scores.view(-1, 1).log()
        match_score += self.match_weights['iou'] * ious
        match_score += self.match_weights['det_label'] * label_deltas

        return match_score

    def assign_ids(self, match_scores: Tensor):
        num_prev_bboxes = match_scores.shape[1] - 1
        _, match_ids = match_scores.max(dim=1)

        ids = match_ids.new_zeros(match_ids.shape[0]) - 1
        best_match_scores = match_scores.new_zeros(num_prev_bboxes) - 1e6
        for idx, match_id in enumerate(match_ids):
            if match_id == 0:
                ids[idx] = self.num_tracks
                self.num_tracks += 1
            else:
                match_score = match_scores[idx, match_id]
                # TODO: fix the bug where multiple candidate might match
                # with the same previous object.
                if match_score > best_match_scores[match_id - 1]:
                    ids[idx] = self.ids[match_id - 1]
                    best_match_scores[match_id - 1] = match_score
        return ids, best_match_scores

    def track(self,
              model: torch.nn.Module,
              img: torch.Tensor,
              feats: List[torch.Tensor],
              data_sample: TrackDataSample,
              rescale=True,
              **kwargs) -> InstanceData:
        """Tracking forward function.

        Args:
            model (nn.Module): VIS model.
            img (Tensor): of shape (T, C, H, W) encoding input image.
                Typically these should be mean centered and std scaled.
                The T denotes the number of key images and usually is 1 in
                MaskTrackRCNN method.
            feats (list[Tensor]): Multi level feature maps of `img`.
            data_sample (:obj:`TrackDataSample`): The data sample.
                It includes information such as `pred_det_instances`.
            rescale (bool, optional): If True, the bounding boxes should be
                rescaled to fit the original scale of the image. Defaults to
                True.

        Returns:
            :obj:`InstanceData`: Tracking results of the input images.
            Each InstanceData usually contains ``bboxes``, ``labels``,
            ``scores`` and ``instances_id``.
        """
        metainfo = data_sample.metainfo
        bboxes = data_sample.pred_det_instances.bboxes
        masks = data_sample.pred_det_instances.masks
        labels = data_sample.pred_det_instances.labels
        scores = data_sample.pred_det_instances.scores

        frame_id = metainfo.get('frame_id', -1)
        # create pred_track_instances
        pred_track_instances = InstanceData()

        if bboxes.shape[0] == 0:
            ids = torch.zeros_like(labels)
            pred_track_instances = data_sample.pred_det_instances.clone()
            pred_track_instances.instances_id = ids
            return pred_track_instances

        rescaled_bboxes = bboxes.clone()
        if rescale:
            scale_factor = rescaled_bboxes.new_tensor(
                metainfo['scale_factor']).repeat((1, 2))
            rescaled_bboxes = rescaled_bboxes * scale_factor
        roi_feats, _ = model.track_head.extract_roi_feats(
            feats, [rescaled_bboxes])

        if self.empty:
            num_new_tracks = bboxes.size(0)
            ids = torch.arange(
                self.num_tracks,
                self.num_tracks + num_new_tracks,
                dtype=torch.long)
            self.num_tracks += num_new_tracks
        else:
            prev_bboxes = self.get('bboxes')
            prev_labels = self.get('labels')
            prev_roi_feats = self.get('roi_feats')

            similarity_logits = model.track_head.predict(
                roi_feats, prev_roi_feats)
            match_scores = self.get_match_score(bboxes, labels, scores,
                                                prev_bboxes, prev_labels,
                                                similarity_logits)
            ids, best_match_scores = self.assign_ids(match_scores)

        valid_inds = ids > -1
        ids = ids[valid_inds]
        bboxes = bboxes[valid_inds]
        labels = labels[valid_inds]
        scores = scores[valid_inds]
        masks = masks[valid_inds]
        roi_feats = roi_feats[valid_inds]

        self.update(
            ids=ids,
            bboxes=bboxes,
            labels=labels,
            scores=scores,
            masks=masks,
            roi_feats=roi_feats,
            frame_ids=frame_id)
        # update pred_track_instances
        pred_track_instances.bboxes = bboxes
        pred_track_instances.masks = masks
        pred_track_instances.labels = labels
        pred_track_instances.scores = scores
        pred_track_instances.instances_id = ids

        return pred_track_instances



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import List, Tuple

from mmdet.datasets.api_wrappers import COCO
from mmengine.fileio import FileClient

from mmtrack.registry import DATASETS
from .api_wrappers import CocoVID
from .base_video_dataset import BaseVideoDataset


@DATASETS.register_module()
class TaoDataset(BaseVideoDataset):
    """Dataset for TAO."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from an annotation file named as ``self.ann_file``

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        if self.load_as_video:
            data_list, valid_data_indices = self._load_tao_data_list()
        else:
            data_list, valid_data_indices = self._load_lvis_data_list()

        return data_list, valid_data_indices

    def _load_tao_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from an annotation file named as ``self.ann_file``

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_path:
            coco = CocoVID(local_path)
        self._metainfo['categories'] = coco.cats
        # The order of returned `cat_ids` will not
        # change with the order of the CLASSES
        self.cat_ids = coco.get_cat_ids(cat_names=self.metainfo['CLASSES'])
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.cat_img_map = copy.deepcopy(coco.cat_img_map)

        data_list = []
        vid_ids = coco.get_vid_ids()
        for vid_id in vid_ids:
            img_ids = coco.get_img_ids_from_vid(vid_id)
            for img_id in img_ids:
                # load img info
                raw_img_info = coco.load_imgs([img_id])[0]
                if raw_img_info['file_name'].startswith('COCO'):
                    # Convert form the COCO 2014 file naming convention of
                    # COCO_[train/val/test]2014_000000000000.jpg to the 2017
                    # naming convention of 000000000000.jpg
                    # (LVIS v1 will fix this naming issue)
                    raw_img_info['filename'] = raw_img_info['file_name'][-16:]
                else:
                    raw_img_info['filename'] = raw_img_info['file_name']
                raw_img_info['img_id'] = img_id
                raw_img_info['video_length'] = len(img_ids)

                # load ann info
                ann_ids = coco.get_ann_ids(
                    img_ids=[img_id], cat_ids=self.cat_ids)
                raw_ann_info = coco.load_anns(ann_ids)

                # get data_info
                parsed_data_info = self.parse_data_info(
                    dict(raw_img_info=raw_img_info, raw_ann_info=raw_ann_info))
                data_list.append(parsed_data_info)

        valid_data_indices = list(range(len(data_list)))
        return data_list, valid_data_indices

    def _load_lvis_data_list(self):
        """Load annotations from an annotation file named as ``self.ann_file``

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_path:
            coco = COCO(local_path)
        # The order of returned `cat_ids` will not
        # change with the order of the CLASSES
        self.cat_ids = coco.get_cat_ids(cat_names=self.metainfo['CLASSES'])
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.cat_img_map = copy.deepcopy(coco.cat_img_map)

        img_ids = coco.get_img_ids()
        data_list = []
        total_ann_ids = []
        for img_id in img_ids:
            raw_img_info = coco.load_imgs([img_id])[0]
            raw_img_info['img_id'] = img_id
            if raw_img_info['file_name'].startswith('COCO'):
                # Convert form the COCO 2014 file naming convention of
                # COCO_[train/val/test]2014_000000000000.jpg to the 2017
                # naming convention of 000000000000.jpg
                # (LVIS v1 will fix this naming issue)
                raw_img_info['filename'] = raw_img_info['file_name'][-16:]
            else:
                raw_img_info['filename'] = raw_img_info['file_name']

            ann_ids = coco.get_ann_ids(img_ids=[img_id])
            raw_ann_info = coco.load_anns(ann_ids)
            total_ann_ids.extend(ann_ids)

            parsed_data_info = self.parse_data_info(
                dict(raw_img_info=raw_img_info, raw_ann_info=raw_ann_info))
            data_list.append(parsed_data_info)

        assert len(set(total_ann_ids)) == len(
            total_ann_ids
        ), f"Annotation ids in '{self.ann_file}' are not unique!"

        valid_data_indices = list(range(len(data_list)))
        return data_list, valid_data_indices



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch
from torch import Tensor

from mmtrack.registry import MODELS, TASK_UTILS
from mmtrack.utils import OptConfigType, OptMultiConfig, SampleList
from .base import BaseMultiObjectTracker


@MODELS.register_module()
class ByteTrack(BaseMultiObjectTracker):
    """ByteTrack: Multi-Object Tracking by Associating Every Detection Box.

    This multi object tracker is the implementation of `ByteTrack
    <https://arxiv.org/abs/2110.06864>`_.

    Args:
        detector (dict): Configuration of detector. Defaults to None.
        tracker (dict): Configuration of tracker. Defaults to None.
        motion (dict): Configuration of motion. Defaults to None.
        data_preprocessor (dict or ConfigDict, optional): The pre-process
           config of :class:`TrackDataPreprocessor`.  it usually includes,
            ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
        init_cfg (dict or list[dict]): Configuration of initialization.
            Defaults to None.
    """

    def __init__(self,
                 detector: Optional[dict] = None,
                 tracker: Optional[dict] = None,
                 motion: Optional[dict] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super().__init__(data_preprocessor, init_cfg)

        if detector is not None:
            self.detector = MODELS.build(detector)

        if motion is not None:
            self.motion = TASK_UTILS.build(motion)

        if tracker is not None:
            self.tracker = MODELS.build(tracker)

    def loss(self, inputs: Dict[str, Tensor], data_samples: SampleList,
             **kwargs) -> dict:
        """Calculate losses from a batch of inputs and data samples.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size.The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Returns:
            dict: A dictionary of loss components.
        """
        # modify the inputs shape to fit mmdet
        img = inputs['img']
        assert img.size(1) == 1
        # convert 'inputs' shape to (N, C, H, W)
        img = torch.squeeze(img, dim=1)
        return self.detector.loss(img, data_samples, **kwargs)

    def predict(self, inputs: Dict[str, Tensor], data_samples: SampleList,
                **kwargs) -> SampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing.

        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size.The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as `gt_instance`.

        Returns:
            SampleList: Tracking results of the input images.
            Each TrackDataSample usually contains ``pred_det_instances``
            or ``pred_track_instances``.
        """
        img = inputs['img']
        assert img.dim() == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        assert img.size(0) == 1, \
            'Bytetrack inference only support 1 batch size per gpu for now.'
        img = img[0]

        assert len(data_samples) == 1, \
            'Bytetrack inference only support 1 batch size per gpu for now.'

        track_data_sample = data_samples[0]

        det_results = self.detector.predict(img, data_samples)
        assert len(det_results) == 1, 'Batch inference is not supported.'
        track_data_sample.pred_det_instances = \
            det_results[0].pred_instances.clone()

        pred_track_instances = self.tracker.track(
            model=self,
            img=img,
            feats=None,
            data_sample=track_data_sample,
            **kwargs)
        track_data_sample.pred_track_instances = pred_track_instances

        return [track_data_sample]



# Copyright (c) OpenMMLab. All rights reserved.
from .aflink import AppearanceFreeLink
from .correlation import depthwise_correlation
from .interpolation import InterpolateTracklets
from .similarity import embed_similarity

__all__ = [
    'depthwise_correlation', 'embed_similarity', 'InterpolateTracklets',
    'AppearanceFreeLink'
]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp

import numpy as np
from mmengine.config import Config, DictAction
from mmengine.dist import get_dist_info
from mmengine.logging import print_log
from mmengine.runner import Runner

from mmtrack.utils import register_all_modules


def parse_range(range_str):
    range_list = range_str.split(',')
    assert len(range_list) == 3 and float(range_list[1]) >= float(
        range_list[0])
    param = map(float, range_list)
    return np.round(np.arange(*param), decimals=2)


def parse_args():
    parser = argparse.ArgumentParser(
        description='MMTrack test (and eval) a model')
    parser.add_argument('config', help='test config file path')
    parser.add_argument('--checkpoint', help='checkpoint file')
    parser.add_argument(
        '--work-dir',
        help='the directory to save the file containing evaluation metrics')
    parser.add_argument(
        '--penalty-k-range',
        type=parse_range,
        help="the range of hyper-parameter 'penalty_k' in SiamRPN++; the format \
            is 'start,stop,step'")
    parser.add_argument(
        '--lr-range',
        type=parse_range,
        help="the range of hyper-parameter 'lr' in SiamRPN++; the format is \
            'start,stop,step'")
    parser.add_argument(
        '--win-influ-range',
        type=parse_range,
        help="the range of hyper-parameter 'window_influence' in SiamRPN++; the \
            format is 'start,stop,step'")
    parser.add_argument(
        '--cfg-options',
        nargs='+',
        action=DictAction,
        help='override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. If the value to '
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        'Note that the quotation marks are necessary and that no white space '
        'is allowed.')
    parser.add_argument(
        '--launcher',
        choices=['none', 'pytorch', 'slurm', 'mpi'],
        default='none',
        help='job launcher')
    parser.add_argument('--local_rank', type=int, default=0)
    args = parser.parse_args()
    if 'LOCAL_RANK' not in os.environ:
        os.environ['LOCAL_RANK'] = str(args.local_rank)
    return args


def parameter_search(runner, args):
    cfg = runner.cfg
    logger = runner.logger

    # calculate the number of all search cases and set comparing standard.
    num_cases = len(args.penalty_k_range) * len(args.lr_range) * len(
        args.win_influ_range)
    case_count = 0
    # compare function setting in parameter search. Now, the default comparing
    # ruler is  `greater` because the model doesn't record comparing ruler
    # of metrics in ``MMEngine``.
    rule_map = {'greater': lambda x, y: x > y, 'less': lambda x, y: x < y}
    compare_func = rule_map['greater']

    if cfg.test_evaluator.metric == 'OPE':
        eval_metrics = ['success', 'norm_precision', 'precision']
        key_metric = 'success'
    else:
        eval_metrics = ['eao', 'accuracy', 'robustness', 'num_fails']
        key_metric = 'eao'

    checkpoint = runner.load_checkpoint(args.checkpoint)

    # init best_score, best_results and best parames
    if 'meta' in checkpoint and 'hook_msgs' in checkpoint[
            'meta'] and key_metric in checkpoint['meta']['hook_msgs']:
        best_score = checkpoint['meta']['hook_msgs'][key_metric]
    else:
        best_score = 0
    best_result = {f'{key_metric}': best_score}

    best_params = dict(
        penalty_k=cfg.model.test_cfg.rpn.penalty_k,
        lr=cfg.model.test_cfg.rpn.lr,
        win_influ=cfg.model.test_cfg.rpn.window_influence)
    print_log(f'init best score as: {best_score}', logger)
    print_log(f'init best params as: {best_params}', logger)

    for penalty_k in args.penalty_k_range:
        for lr in args.lr_range:
            for win_influ in args.win_influ_range:
                case_count += 1
                runner.model.test_cfg.rpn.penalty_k = penalty_k
                runner.model.test_cfg.rpn.lr = lr
                runner.model.test_cfg.rpn.window_influence = win_influ
                print_log(f'-----------[{case_count}/{num_cases}]-----------',
                          logger)
                print_log(
                    f'penalty_k={penalty_k} lr={lr} win_influence={win_influ}',
                    logger)

                # start testing
                runner.test()

                # parse the eluation results
                res = dict()
                for metric in eval_metrics:
                    res[metric] = runner.message_hub.get_scalar(
                        'test/sot/' + metric).current()

                # show results
                rank, _ = get_dist_info()
                if rank == 0:
                    print_log(f'evaluation results: {res}', logger)
                    print_log('------------------------------------------',
                              logger)
                    if compare_func(res[key_metric], best_result[key_metric]):
                        best_result = res
                        best_params['penalty_k'] = penalty_k
                        best_params['lr'] = lr
                        best_params['win_influ'] = win_influ
                    print_log(
                        f'The current best evaluation results: {best_result}',
                        logger)
                    print_log(f'The current best params: {best_params}',
                              logger)

    print_log(
        'After parameter searching, the best evaluation results: '
        f'{best_result}', logger)
    print_log(f'After parameter searching, the best params: {best_params}',
              logger)


def main():
    args = parse_args()

    # register all modules in mmtrack into the registries
    # do not init the default scope here because it will be init in the runner
    register_all_modules(init_default_scope=False)

    # load config
    cfg = Config.fromfile(args.config)
    cfg.launcher = args.launcher
    if args.cfg_options is not None:
        cfg.merge_from_dict(args.cfg_options)

    # work_dir is determined in this priority: CLI > segment in file > filename
    if args.work_dir is not None:
        # update configs according to CLI args if args.work_dir is not None
        cfg.work_dir = args.work_dir
    elif cfg.get('work_dir', None) is None:
        # use config filename as default work_dir if cfg.work_dir is None
        cfg.work_dir = osp.join('./work_dirs',
                                osp.splitext(osp.basename(args.config))[0])

    cfg.load_from = args.checkpoint

    # build the runner from config
    runner = Runner.from_cfg(cfg)

    parameter_search(runner, args)


if __name__ == '__main__':
    main()



_base_ = [
    './tracktor_faster-rcnn_r50-fpn_8xb2-4e_mot17halftrain'
    '_test-mot17halfval.py'
]

model = dict(
    detector=dict(
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/faster_rcnn/faster-rcnn_r50_fpn_4e_mot15-half_20210804_001040-ae733d0c.pth'  # noqa: E501
        )),
    reid=dict(
        head=dict(num_classes=375),
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/mot/reid/reid_r50_6e_mot15_20210803_192157-65b5e2d7.pth'  # noqa: E501
        )))

# dataloader
data_root = 'data/MOT15/'
val_dataloader = dict(
    dataset=dict(
        data_root=data_root,
        ann_file='annotations/half-val_cocoformat.json',
    ))
test_dataloader = val_dataloader



# Copyright (c) OpenMMLab. All rights reserved.
import random
from typing import List, Optional

from mmengine.dataset import ConcatDataset

from mmtrack.registry import DATASETS


@DATASETS.register_module()
class RandomSampleConcatDataset(ConcatDataset):
    """A wrapper of concatenated dataset. Support randomly sampling one dataset
    from concatenated datasets and then getting samples from the sampled
    dataset. This class only support training.

    Args:
        datasets (list[dict]): The list contains all configs of
            concatenated datasets.
        dataset_sampling_weights (Optional[List[float]], optional): The list
            contains the sampling weights of each dataset. Defaults to None.
    """

    def __init__(self,
                 datasets: List[dict],
                 dataset_sampling_weights: Optional[List[float]] = None):
        if dataset_sampling_weights is None:
            self.dataset_sampling_probs = [1. / len(datasets)] * len(datasets)
        else:
            for x in dataset_sampling_weights:
                assert x >= 0.
            prob_total = float(sum(dataset_sampling_weights))
            assert prob_total > 0.
            self.dataset_sampling_probs = [
                x / prob_total for x in dataset_sampling_weights
            ]

        datasets = [DATASETS.build(cfg) for cfg in datasets]
        # add an attribute `CLASSES` for the calling in `tools/train.py`
        self.CLASSES = datasets[0].META['CLASSES']

        super().__init__(datasets)

    def __getitem__(self, ind: int) -> dict:
        """Random sampling a dataset and get samples from this dataset..

        Args:
            ind (int): The random index.  Actually, in this class,
                the input 'ind' is not used in 'dataset'.

        Returns:
            dict: The results after the dataset pipeline.
        """
        while True:
            dataset = random.choices(self.datasets,
                                     self.dataset_sampling_probs)[0]
            ind = random.randint(0, len(dataset) - 1)
            results = dataset[ind]
            if results is not None:
                return results



_base_ = [
    '../../_base_/models/faster-rcnn_r50-dc5.py',
    '../../_base_/datasets/imagenet_vid_fgfa_style.py',
    '../../_base_/default_runtime.py'
]
model = dict(
    type='SELSA',
    detector=dict(
        roi_head=dict(
            type='mmtrack.SelsaRoIHead',
            bbox_head=dict(
                type='mmtrack.SelsaBBoxHead',
                num_shared_fcs=2,
                aggregator=dict(
                    type='mmtrack.SelsaAggregator',
                    in_channels=1024,
                    num_attention_blocks=16)),
            bbox_roi_extractor=dict(
                type='mmtrack.SingleRoIExtractor',
                roi_layer=dict(
                    type='RoIAlign', output_size=7, sampling_ratio=2),
                out_channels=512,
                featmap_strides=[16]))))

# dataset settings
val_dataloader = dict(
    dataset=dict(
        ref_img_sampler=dict(
            _delete_=True,
            num_ref_imgs=14,
            frame_range=[-7, 7],
            method='test_with_adaptive_stride')))
test_dataloader = val_dataloader

# training schedule
train_cfg = dict(type='EpochBasedTrainLoop', max_epochs=7, val_interval=7)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='SGD', lr=0.01, momentum=0.9, weight_decay=0.0001),
    clip_grad=dict(max_norm=35, norm_type=2))

# learning rate
param_scheduler = [
    dict(
        type='LinearLR',
        start_factor=1.0 / 3,
        by_epoch=False,
        begin=0,
        end=500),
    dict(
        type='MultiStepLR',
        begin=0,
        end=7,
        by_epoch=True,
        milestones=[2, 5],
        gamma=0.1)
]

visualizer = dict(type='DetLocalVisualizer')



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
import time
from typing import List

import numpy as np

from mmtrack.registry import DATASETS
from .base_sot_dataset import BaseSOTDataset


@DATASETS.register_module()
class GOT10kDataset(BaseSOTDataset):
    """GOT10k Dataset of single object tracking.

    The dataset can both support training and testing mode.
    """

    def __init__(self, *args, **kwargs):
        """Initialization of SOT dataset class."""
        super(GOT10kDataset, self).__init__(*args, **kwargs)

    def load_data_list(self) -> List[dict]:
        """Load annotations from an annotation file named as ``self.ann_file``.

        Returns:
            list[dict]: the length of the list is the number of videos. The
                inner dict is in the following format:
                    {
                        'video_path': the video path
                        'ann_path': the annotation path
                        'start_frame_id': the starting frame number contained
                            in the image name
                        'end_frame_id': the ending frame number contained in
                            the image name
                        'framename_template': the template of image name
                    }
        """
        print('Loading GOT10k dataset...')
        start_time = time.time()
        data_infos = []
        data_infos_str = self._loadtxt(
            self.ann_file, return_ndarray=False).split('\n')
        # the first line of annotation file is a dataset comment.
        for line in data_infos_str[1:]:
            # compatible with different OS.
            line = line.strip().replace('/', os.sep).split(',')
            data_info = dict(
                video_path=line[0],
                ann_path=line[1],
                start_frame_id=int(line[2]),
                end_frame_id=int(line[3]),
                framename_template='%08d.jpg')
            data_infos.append(data_info)
        print(f'GOT10k dataset loaded! ({time.time()-start_time:.2f} s)')
        return data_infos

    def get_visibility_from_video(self, video_idx: int) -> dict:
        """Get the visible information of instance in a video.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: The visibilities of each object in the video.
        """
        if not self.test_mode:
            video_path = self.get_data_info(video_idx)['video_path']
            absense_info_path = osp.join(self.data_prefix['img_path'],
                                         video_path, 'absence.label')
            cover_info_path = osp.join(self.data_prefix['img_path'],
                                       video_path, 'cover.label')
            absense_info = self._loadtxt(absense_info_path, dtype=bool)
            # The values of key 'cover' are
            # int numbers in range [0,8], which correspond to
            # ranges of object visible ratios: 0%, (0%, 15%],
            # (15%~30%], (30%, 45%], (45%, 60%],(60%, 75%],
            # (75%, 90%], (90%, 100%) and 100% respectively
            cover_info = self._loadtxt(cover_info_path, dtype=int)
            visible = ~absense_info & (cover_info > 0)
            visible_ratio = cover_info / 8.
            return dict(visible=visible, visible_ratio=visible_ratio)
        else:
            return super(GOT10kDataset,
                         self).get_visibility_from_video(video_idx)

    def prepare_test_data(self, video_idx: int, frame_idx: int) -> dict:
        """Get testing data of one frame. We parse one video, get one frame
        from it and pass the frame information to the pipeline.

        Args:
            video_idx (int): The index of video.
            frame_idx (int): The index of frame.

        Returns:
            dict: Testing data of one frame.
        """
        if self.test_memo.get('video_idx', None) != video_idx:
            self.test_memo.video_idx = video_idx
            ann_infos = self.get_ann_infos_from_video(video_idx)
            img_infos = self.get_img_infos_from_video(video_idx)
            self.test_memo.video_infos = dict(**img_infos, **ann_infos)
        assert 'video_idx' in self.test_memo and 'video_infos'\
            in self.test_memo

        results = {}
        results['img_path'] = self.test_memo.video_infos['img_paths'][
            frame_idx]
        results['frame_id'] = frame_idx
        results['video_id'] = video_idx
        results['video_length'] = self.test_memo.video_infos['video_length']

        instance = {}
        if frame_idx == 0:
            ann_infos = self.get_ann_infos_from_video(video_idx)
            instance['bbox'] = ann_infos['bboxes'][frame_idx]

        results['instances'] = []
        instance['visible'] = True
        instance['bbox_label'] = np.array([0], dtype=np.int32)
        results['instances'].append(instance)
        results = self.pipeline(results)
        return results



_base_ = ['../../_base_/default_runtime.py']

randomness = dict(seed=1, deterministic=True)
find_unused_parameters = True
crop_size = 511
exemplar_size = 127
search_size = 255

# model settings
model = dict(
    type='SiamRPN',
    data_preprocessor=dict(type='TrackDataPreprocessor'),
    backbone=dict(
        type='SOTResNet',
        depth=50,
        out_indices=(1, 2, 3),
        frozen_stages=4,
        strides=(1, 2, 1, 1),
        dilations=(1, 1, 2, 4),
        norm_eval=True,
        init_cfg=dict(
            type='Pretrained',
            checkpoint=  # noqa: E251
            'https://download.openmmlab.com/mmtracking/pretrained_weights/sot_resnet50.model'  # noqa: E501
        )),
    neck=dict(
        type='mmdet.ChannelMapper',
        in_channels=[512, 1024, 2048],
        out_channels=256,
        kernel_size=1,
        norm_cfg=dict(type='BN'),
        act_cfg=None),
    head=dict(
        type='SiameseRPNHead',
        anchor_generator=dict(
            type='SiameseRPNAnchorGenerator',
            strides=[8],
            ratios=[0.33, 0.5, 1, 2, 3],
            scales=[8]),
        in_channels=[256, 256, 256],
        weighted_sum=True,
        bbox_coder=dict(
            type='mmdet.DeltaXYWHBBoxCoder',
            target_means=[0., 0., 0., 0.],
            target_stds=[1., 1., 1., 1.]),
        loss_cls=dict(
            type='mmdet.CrossEntropyLoss', reduction='sum', loss_weight=1.0),
        loss_bbox=dict(type='mmdet.L1Loss', reduction='sum', loss_weight=1.2)),
    train_cfg=dict(
        rpn=dict(
            assigner=dict(
                type='mmdet.MaxIoUAssigner',
                pos_iou_thr=0.6,
                neg_iou_thr=0.3,
                min_pos_iou=0.6,
                match_low_quality=False,
                iou_calculator=dict(type='mmdet.BboxOverlaps2D')),
            sampler=dict(
                type='mmdet.RandomSampler',
                num=64,
                pos_fraction=0.25,
                add_gt_as_proposals=False),
            num_neg=16,
            exemplar_size=exemplar_size,
            search_size=search_size)),
    test_cfg=dict(
        exemplar_size=exemplar_size,
        search_size=search_size,
        context_amount=0.5,
        center_size=7,
        rpn=dict(penalty_k=0.05, window_influence=0.42, lr=0.38)))

# data pipeline
data_root = 'data/'
train_pipeline = [
    dict(
        type='PairSampling',
        frame_range=100,
        pos_prob=0.8,
        filter_template_img=False),
    dict(
        type='TransformBroadcaster',
        share_random_params=False,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=False),
            dict(
                type='CropLikeSiamFC',
                context_amount=0.5,
                exemplar_size=exemplar_size,
                crop_size=crop_size)
        ]),
    dict(
        type='SeqShiftScaleAug',
        target_size=[exemplar_size, search_size],
        shift=[4, 64],
        scale=[0.05, 0.18]),
    dict(type='SeqColorAug', prob=[1.0, 1.0]),
    dict(type='SeqBlurAug', prob=[0.0, 0.2]),
    dict(type='PackTrackInputs', ref_prefix='search', num_template_frames=1)
]

# dataloader
train_dataloader = dict(
    batch_size=28,
    num_workers=4,
    persistent_workers=True,
    sampler=dict(type='QuotaSampler', samples_per_epoch=600000),
    dataset=dict(
        type='RandomSampleConcatDataset',
        dataset_sampling_weights=[0.25, 0.2, 0.55],
        datasets=[
            dict(
                type='SOTImageNetVIDDataset',
                data_root=data_root,
                ann_file='ILSVRC/annotations/imagenet_vid_train.json',
                data_prefix=dict(img_path='ILSVRC/Data/VID'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='SOTCocoDataset',
                data_root=data_root,
                ann_file='coco/annotations/instances_train2017.json',
                data_prefix=dict(img_path='coco/train2017'),
                pipeline=train_pipeline,
                test_mode=False),
            dict(
                type='SOTCocoDataset',
                data_root=data_root,
                ann_file='ILSVRC/annotations/imagenet_det_30plus1cls.json',
                data_prefix=dict(img_path='ILSVRC/Data/DET'),
                pipeline=train_pipeline,
                test_mode=False)
        ]))

# runner loop
train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=20, val_begin=10, val_interval=1)

# learning policy
param_scheduler = [
    dict(
        type='SiamRPNExpLR',
        start_factor=0.2,
        end_factor=1.0,
        by_epoch=True,
        begin=0,
        end=5,
        endpoint=False),
    dict(
        type='SiamRPNExpLR',
        start_factor=1.0,
        end_factor=0.1,
        by_epoch=True,
        begin=5,
        end=20,
        endpoint=True)
]

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='SGD', lr=0.005, momentum=0.9, weight_decay=0.0001),
    clip_grad=dict(max_norm=10.0, norm_type=2),
    paramwise_cfg=dict(
        custom_keys=dict(backbone=dict(lr_mult=0.1, decay_mult=1.0))))

custom_hooks = [
    dict(
        type='SiamRPNBackboneUnfreezeHook',
        backbone_start_train_epoch=10,
        backbone_train_layers=['layer2', 'layer3', 'layer4'])
]



_base_ = [
    './faster-rcnn_resnet50-fpn_8xb2-4e_mot17halftrain_test-mot17halfval.py'
]
# data
data_root = 'data/MOT15/'
train_dataloader = dict(dataset=dict(data_root=data_root))
val_dataloader = dict(dataset=dict(data_root=data_root))
test_dataloader = val_dataloader

val_evaluator = dict(ann_file=data_root +
                     'annotations/half-val_cocoformat.json')
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import tempfile
from collections import defaultdict
from typing import Dict, List, Optional, Sequence, Tuple, Union

import mmengine
import numpy as np
from mmengine.logging import MMLogger

from mmtrack.registry import METRICS
from .base_video_metrics import BaseVideoMetric

try:
    import tao
    from tao.toolkit.tao import Tao, TaoEval
except ImportError:
    tao = None

try:
    import lvis
    from lvis import LVIS, LVISEval, LVISResults
except ImportError:
    lvis = None


@METRICS.register_module()
class TAOMetric(BaseVideoMetric):
    """mAP evaluation metrics for the TAO task.

    Args:
        metric (str | list[str]): Metrics to be evaluated.
            Defaults to 'tao_track_ap'.
        metric_items (List[str], optional): Metric result names to be
            recorded in the evaluation result. Defaults to None.
        outfile_prefix (str | None): The prefix of json files. It includes
            the file path and the prefix of filename, e.g., "a/b/prefix".
            If not specified, a temp file will be created. Defaults to None.
        collect_device (str): Device name used for collecting results from
            different ranks during distributed training. Must be 'cpu' or
            'gpu'. Defaults to 'cpu'.
        prefix (str, optional): The prefix that will be added in the metric
            names to disambiguate homonyms metrics of different evaluators.
            If prefix is not provided in the argument, self.default_prefix
            will be used instead. Defaults to None.
        format_only (bool): If True, only formatting the results to the
            official format and not performing evaluation. Defaults to False.
    """

    default_prefix: Optional[str] = 'tao'

    def __init__(self,
                 metric: Union[str, List[str]] = 'tao_track_ap',
                 metric_items: Optional[Sequence[str]] = None,
                 outfile_prefix: Optional[str] = None,
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None,
                 format_only: bool = False) -> None:
        super().__init__(collect_device=collect_device, prefix=prefix)
        # tao evaluation metrics
        self.metrics = metric if isinstance(metric, list) else [metric]
        self.format_only = format_only
        allowed_metrics = ['tao_track_ap', 'bbox']
        for metric in self.metrics:
            if metric not in allowed_metrics:
                raise KeyError(f"metric should be 'tao_track_ap' or 'bbox',"
                               f' but got {metric}.')

        self.metric_items = metric_items
        self.outfile_prefix = outfile_prefix
        self.per_video_res = []
        self.img_ids = []

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.results``, which will be used to
        compute the metrics when all batches have been processed.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of data samples that
                contain annotations and predictions.
        """
        for data_sample in data_samples:
            result = dict()
            pred_track = data_sample['pred_track_instances']
            pred_det = data_sample['pred_det_instances']
            frame_id = data_sample['frame_id']
            video_length = data_sample['video_length']

            result['img_id'] = data_sample['img_id']
            result['track_bboxes'] = pred_track['bboxes'].cpu().numpy()
            result['track_scores'] = pred_track['scores'].cpu().numpy()
            result['track_labels'] = pred_track['labels'].cpu().numpy()
            result['track_instances_id'] = pred_track['instances_id'].cpu(
            ).numpy()

            result['det_bboxes'] = pred_det['bboxes'].cpu().numpy()
            result['det_scores'] = pred_det['scores'].cpu().numpy()
            result['det_labels'] = pred_det['labels'].cpu().numpy()

            # parse gt
            gt = dict()
            gt['width'] = data_sample['ori_shape'][1]
            gt['height'] = data_sample['ori_shape'][0]
            keys = [
                'frame_id', 'frame_index', 'neg_category_ids',
                'not_exhaustive_category_ids', 'img_id', 'video_id',
                'video_length'
            ]
            for key in keys:
                if key not in data_sample:
                    raise KeyError(
                        f'The key {key} is not found in track_data_sample,'
                        f' please pass it into the meta_keys'
                        f' of the PackTrackInputs')
                gt[key] = data_sample[key]

            # When the ground truth exists, get annotation from `instances`.
            # In general, it contains `bbox`, `bbox_label` and `instance_id`.
            if 'instances' in data_sample:
                gt['anns'] = data_sample['instances']
            else:
                gt['anns'] = dict()
            self.per_video_res.append((result, gt))

            if frame_id == video_length - 1:
                preds, gts = zip(*self.per_video_res)
                # format the results
                gt_results, tao_meta_info = self._format_one_video_gts(gts)
                pred_track_results, pred_det_results = \
                    self._format_one_video_preds(preds, tao_meta_info)
                self.per_video_res.clear()
                # add converted result to the results list
                self.results.append((pred_track_results, pred_det_results,
                                     gt_results, tao_meta_info))

    def compute_metrics(self, results: List) -> Dict[str, float]:
        """Compute the metrics from processed results.

        Args:
            results (List): The processed results of each batch.

        Returns:
            Dict[str, float]: The computed metrics. The keys are the names of
            the metrics, and the values are corresponding results.
        """
        logger: MMLogger = MMLogger.get_current_instance()
        # split gt and prediction list
        tmp_pred_track_results, tmp_pred_det_results, \
            tmp_gt_results, tmp_meta_info = zip(*results)
        tao_meta_info = self.format_meta(tmp_meta_info)
        gt_results = self.format_gts(tmp_gt_results, tao_meta_info)
        pred_track_results = self.format_preds(tmp_pred_track_results)
        pred_det_results = self.format_preds(tmp_pred_det_results)

        if 'bbox' in self.metrics:
            # LVIS api only supports reading from files, hence,
            # save the json result to tmp dir
            tmp_dir = tempfile.TemporaryDirectory()
            pred_det_results_path = f'{tmp_dir.name}/tao_bbox.json'
            gt_results_path = f'{tmp_dir.name}/tao_gt.json'
            mmengine.dump(pred_det_results, pred_det_results_path)
            mmengine.dump(gt_results, gt_results_path)

        if self.format_only:
            self.save_pred_results(pred_track_results, 'track')
            self.save_pred_results(pred_det_results, 'det')
            return dict()

        eval_results = dict()

        if 'tao_track_ap' in self.metrics:
            if tao is None:
                raise ImportError(
                    'Please run'
                    ' pip install git+https://github.com/TAO-Dataset/tao.git '
                    'to manually install tao')

            logger.info('Evaluating tracking results...')
            tao_gt = Tao(gt_results)
            tao_eval = TaoEval(tao_gt, pred_track_results)
            tao_eval.params.img_ids = self.img_ids
            tao_eval.params.cat_ids = list(
                self.dataset_meta['categories'].keys())
            tao_eval.params.iou_thrs = np.array([0.5, 0.75])
            tao_eval.run()

            tao_eval.print_results()
            tao_results = tao_eval.get_results()
            for k, v in tao_results.items():
                if isinstance(k, str) and k.startswith('AP'):
                    key = 'track_{}'.format(k)
                    val = float('{:.3f}'.format(float(v)))
                    eval_results[key] = val

        if 'bbox' in self.metrics:
            if lvis is None:
                raise ImportError(
                    'Please run'
                    ' pip install git+https://github.com/lvis-dataset/lvis-api.git '  # noqa
                    'to manually install lvis')

            logger.info('Evaluating detection results...')
            lvis_gt = LVIS(gt_results_path)
            lvis_dt = LVISResults(lvis_gt, pred_det_results_path)
            lvis_eval = LVISEval(lvis_gt, lvis_dt, 'bbox')
            lvis_eval.params.imgIds = self.img_ids
            lvis_eval.params.catIds = list(
                self.dataset_meta['categories'].keys())
            lvis_eval.evaluate()
            lvis_eval.accumulate()
            lvis_eval.summarize()
            lvis_eval.print_results()

            lvis_results = lvis_eval.get_results()
            for k, v in lvis_results.items():
                if k.startswith('AP'):
                    key = '{}_{}'.format('bbox', k)
                    val = float('{:.3f}'.format(float(v)))
                    eval_results[key] = val
            tmp_dir.cleanup()
        return eval_results

    def format_meta(self, parts_meta: Tuple[dict]) -> dict:
        """Gather all meta info from self.results."""
        all_seq_vids_info = []
        all_seq_imgs_info = []
        all_seq_tracks_info = []
        for _seq_info in parts_meta:
            all_seq_vids_info.extend(_seq_info['videos'])
            all_seq_imgs_info.extend(_seq_info['images'])
            all_seq_tracks_info.extend(_seq_info['tracks'])

        # update tao_meta_info
        tao_meta_info = dict(
            videos=all_seq_vids_info,
            images=all_seq_imgs_info,
            tracks=all_seq_tracks_info)

        return tao_meta_info

    def format_gts(self, gts: Tuple[List], tao_meta_info: dict) -> dict:
        """Gather all ground-truth from self.results."""
        categories = list(self.dataset_meta['categories'].values())
        for img_info in tao_meta_info['images']:
            self.img_ids.append(img_info['id'])
        gt_results = dict(
            info=dict(),
            images=tao_meta_info['images'],
            categories=categories,
            videos=tao_meta_info['videos'],
            annotations=[],
            tracks=tao_meta_info['tracks'])

        ann_id = 1
        for gt_result in gts:
            for ann in gt_result:
                ann['id'] = ann_id
                gt_results['annotations'].append(ann)
                ann_id += 1

        return gt_results

    def format_preds(self, preds: Tuple[List]) -> List:
        """Gather all predictions from self.results."""
        pred_results = []
        max_track_id = 0
        for pred_result in preds:
            # update track id
            if 'track_id' in pred_result[0]:
                track_ids = []
                for ins_info in pred_result:
                    track_ids.append(ins_info['track_id'])
                    ins_info['track_id'] += max_track_id
                track_ids = list(set(track_ids))
                max_track_id += max(track_ids) + 1

            pred_results.extend(pred_result)
        return pred_results

    def _format_one_video_preds(self, pred_dicts: Tuple[dict],
                                tao_meta_info: Dict) -> Tuple[List, List]:
        """Convert the annotation to the format of YouTube-VIS.

        This operation is to make it easier to use the official eval API.

        Args:
            pred_dicts (Tuple[dict]): Prediction of the dataset.
            tao_meta_info (dict): A dict containing videos and images
                information of TAO.

        Returns:
            List: The formatted predictions.
        """
        # Collate preds scatters (tuple of dict to dict of list)
        preds = defaultdict(list)
        cat_ids = list(self.dataset_meta['categories'].keys())
        for pred in pred_dicts:
            for key in pred.keys():
                preds[key].append(pred[key])

        vid_infos = tao_meta_info['videos']
        track_json_results = []
        det_json_results = []
        video_id = vid_infos[-1]['id']

        for img_id, bboxes, scores, labels, ins_ids in zip(
                preds['img_id'], preds['track_bboxes'], preds['track_scores'],
                preds['track_labels'], preds['track_instances_id']):
            for bbox, score, label, ins_id in zip(bboxes, scores, labels,
                                                  ins_ids):
                data = dict(
                    image_id=img_id,
                    bbox=[
                        bbox[0],
                        bbox[1],
                        bbox[2] - bbox[0],
                        bbox[3] - bbox[1],
                    ],
                    score=score,
                    category_id=cat_ids[label],
                    track_id=ins_id,
                    video_id=video_id)
                track_json_results.append(data)

        for img_id, bboxes, scores, labels in zip(preds['img_id'],
                                                  preds['det_bboxes'],
                                                  preds['det_scores'],
                                                  preds['det_labels']):
            for bbox, score, label in zip(bboxes, scores, labels):
                data = dict(
                    image_id=img_id,
                    bbox=[
                        bbox[0],
                        bbox[1],
                        bbox[2] - bbox[0],
                        bbox[3] - bbox[1],
                    ],
                    score=score,
                    category_id=cat_ids[label],
                    video_id=video_id)
                det_json_results.append(data)

        return track_json_results, det_json_results

    def _format_one_video_gts(self,
                              gt_dicts: Tuple[dict]) -> Tuple[list, dict]:
        """Convert the annotation to the format of YouTube-VIS.

        This operation is to make it easier to use the official eval API.

        Args:
            gt_dicts (Tuple[dict]): Ground truth of the dataset.

        Returns:
            Tuple[list, dict]: The formatted gts and a dict containing videos
            and images information of TAO.
        """
        video_infos = []
        image_infos = []
        track_infos = []
        annotations = []
        instance_flag = dict()  # flag the ins_id is used or not
        tao_meta_info = defaultdict(list)
        cat_ids = list(self.dataset_meta['categories'].keys())

        # get video infos
        for gt_dict in gt_dicts:
            frame_id = gt_dict['frame_id']
            video_id = gt_dict['video_id']
            img_id = gt_dict['img_id']
            image_info = dict(
                id=img_id,
                width=gt_dict['width'],
                height=gt_dict['height'],
                video_id=video_id,
                frame_id=frame_id,
                frame_index=gt_dict['frame_index'],
                neg_category_ids=gt_dict['neg_category_ids'],
                not_exhaustive_category_ids=gt_dict[
                    'not_exhaustive_category_ids'],
                file_name='')
            image_infos.append(image_info)
            if frame_id == 0:
                video_info = dict(
                    id=video_id,
                    width=gt_dict['width'],
                    height=gt_dict['height'],
                    neg_category_ids=gt_dict['neg_category_ids'],
                    not_exhaustive_category_ids=gt_dict[
                        'not_exhaustive_category_ids'],
                    file_name='')
                video_infos.append(video_info)

            for ann in gt_dict['anns']:
                label = ann['bbox_label']
                bbox = ann['bbox']
                instance_id = ann['instance_id']
                coco_bbox = [
                    bbox[0],
                    bbox[1],
                    bbox[2] - bbox[0],
                    bbox[3] - bbox[1],
                ]

                annotation = dict(
                    id=-1,  # need update when all results have been collected
                    video_id=video_id,
                    image_id=img_id,
                    frame_id=frame_id,
                    bbox=coco_bbox,
                    track_id=instance_id,
                    instance_id=instance_id,
                    iscrowd=ann.get('ignore_flag', 0),
                    category_id=cat_ids[label],
                    area=coco_bbox[2] * coco_bbox[3])
                if not instance_flag.get(instance_id, False):
                    track_info = dict(
                        id=instance_id,
                        category_id=cat_ids[label],
                        video_id=video_id)
                    track_infos.append(track_info)
                    instance_flag[instance_id] = True
                annotations.append(annotation)

        # update tao meta info
        tao_meta_info['images'].extend(image_infos)
        tao_meta_info['videos'].extend(video_infos)
        tao_meta_info['tracks'].extend(track_infos)

        return annotations, tao_meta_info

    def save_pred_results(self, pred_results: List, res_type: str) -> None:
        """Save the results to a zip file.

        Args:
            pred_results (list): Testing results of the
                dataset.
            res_type (str): The type of testing results, track or detection.
        """
        logger: MMLogger = MMLogger.get_current_instance()
        if self.outfile_prefix is None:
            tmp_dir = tempfile.TemporaryDirectory()
            outfile_prefix = osp.join(tmp_dir.name, 'results')
        else:
            outfile_prefix = self.outfile_prefix
        mmengine.dump(pred_results, f'{outfile_prefix}_{res_type}.json')

        logger.info(f'save the results to {outfile_prefix}_{res_type}.json')



_base_ = [
    './qdtrack_faster-rcnn_r50_fpn_8xb2-4e_mot17halftrain_'
    'test-mot17halfval.py'
]

# data pipeline
train_pipeline = [
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=True),
            dict(
                type='mmdet.RandomResize',
                resize_type='mmdet.Resize',
                scale=(1088, 1088),
                ratio_range=(0.8, 1.2),
                keep_ratio=True,
                clip_object_border=False),
            dict(type='mmdet.PhotoMetricDistortion')
        ]),
    dict(
        type='TransformBroadcaster',
        share_random_params=False,
        transforms=[
            dict(
                type='mmdet.RandomCrop',
                crop_size=(1088, 1088),
                bbox_clip_border=False)
        ]),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='mmdet.RandomFlip', prob=0.5),
        ]),
    dict(type='PackTrackInputs', ref_prefix='ref')
]
mot_cfg = dict(
    type='MOTChallengeDataset',
    data_root='data/MOT17',
    metainfo=dict(CLASSES=('pedestrian')),
    visibility_thr=-1,
    ann_file='annotations/half-train_cocoformat.json',
    data_prefix=dict(img_path='train'),
    ref_img_sampler=dict(
        num_ref_imgs=1, frame_range=10, filter_key_img=True, method='uniform'),
    pipeline=train_pipeline)
crowdhuman_cfg = dict(
    type='BaseVideoDataset',
    data_root='data/crowdhuman',
    load_as_video=False,
    metainfo=dict(CLASSES=('pedestrian')),
    ann_file='annotations/crowdhuman_train.json',
    data_prefix=dict(img_path='train'),
    ref_img_sampler=dict(num_ref_imgs=1, frame_range=0),
    pipeline=train_pipeline)

train_dataloader = dict(
    batch_size=2,
    num_workers=2,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    batch_sampler=dict(type='mmdet.AspectRatioBatchSampler'),
    dataset=dict(
        _delete_=True,
        type='ConcatDataset',
        datasets=[mot_cfg, crowdhuman_cfg]))



# Copyright (c) OpenMMLab. All rights reserved.
"""Collecting some commonly used type hint in mmdetection."""
from typing import Dict, List, Optional, Tuple, Union

import torch
from mmengine.config import ConfigDict
from mmengine.structures import InstanceData

from ..structures import TrackDataSample

# Type hint of config data
ConfigType = Union[ConfigDict, dict]
OptConfigType = Optional[ConfigType]
# Type hint of one or more config data
MultiConfig = Union[ConfigType, List[ConfigType]]
OptMultiConfig = Optional[MultiConfig]

InstanceList = List[InstanceData]
OptInstanceList = Optional[InstanceList]

SampleList = List[TrackDataSample]
OptSampleList = Optional[SampleList]

ForwardResults = Union[Dict[str, torch.Tensor], List[TrackDataSample],
                       Tuple[torch.Tensor], torch.Tensor]



# Copyright (c) OpenMMLab. All rights reserved.
from .bbox_heads import SelsaBBoxHead
from .roi_extractors import SingleRoIExtractor
from .selsa_roi_head import SelsaRoIHead

__all__ = ['SelsaRoIHead', 'SelsaBBoxHead', 'SingleRoIExtractor']



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import os.path as osp
import random
from typing import Any, List, Tuple

from mmdet.datasets.api_wrappers import COCO
from mmengine.dataset import BaseDataset, force_full_init
from mmengine.fileio import FileClient
from mmengine.logging import MMLogger

from mmtrack.registry import DATASETS
from .api_wrappers import CocoVID


@DATASETS.register_module()
class BaseVideoDataset(BaseDataset):
    """Base video dataset for VID, MOT and VIS tasks, except for SOT tasks.

    Args:
        load_as_video (bool, optional): Load data as videos or images.
            Defaults to True.
        key_img_sampler (dict, optional): Configuration of sampling key images.
            Defaults to dict(interval=1).
        ref_img_sampler (dict, optional): Configuration of sampling
            reference images.
            - num_ref_imgs (int, optional): The number of sampled reference
                images. Defaults to 2.
            - frame_range (List(int) | int, optional): The sampling range of
                reference frames in the same video for key frame.
                Defaults to 9.
            - filter_key_img (bool, optional): If False, the key image will be
                in the sampling reference candidates, otherwise, it is exclude.
                Defaults to True.
            - method (str, optional): The sampling method. Options are
                'uniform', 'bilateral_uniform', 'test_with_adaptive_stride',
                'test_with_fix_stride'. Defaults to 'bilateral_uniform'.
    """
    META = dict(CLASSES=None)

    def __init__(self,
                 load_as_video: bool = True,
                 key_img_sampler: dict = dict(interval=1),
                 ref_img_sampler: dict = dict(
                     num_ref_imgs=2,
                     frame_range=9,
                     filter_key_img=True,
                     method='bilateral_uniform'),
                 *args,
                 **kwargs):
        self.load_as_video = load_as_video
        self.key_img_sampler = key_img_sampler
        self.ref_img_sampler = ref_img_sampler
        super().__init__(*args, **kwargs)

    def full_init(self):
        """Load annotation file and set ``BaseDataset._fully_initialized`` to
        True.

        If ``lazy_init=False``, ``full_init`` will be called during the
        instantiation and ``self._fully_initialized`` will be set to True. If
        ``obj._fully_initialized=False``, the class method decorated by
        ``force_full_init`` will call ``full_init`` automatically.

        Several steps to initialize annotation:

            - load_data_list: Load annotations from annotation file.
            - filter data information: Filter annotations according to
              filter_cfg.
            - serialize_data: Serialize ``self.data_list`` if
            ``self.serialize_data`` is True.
        """
        if self._fully_initialized:
            return
        # Load data information.
        # We use `self.valid_data_inds` to record the ids of `data_list` used
        # for training and testing.
        self.data_list, self.valid_data_indices = self.load_data_list()
        # Filter illegal data, such as data that has no annotations.
        self.valid_data_indices = self.filter_data()

        # Serialize data_list
        if self.serialize_data:
            self.data_bytes, self.data_address = self._serialize_data()

        self._fully_initialized = True

    def load_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from an annotation file named as ``self.ann_file``.
        Specifically, if self.load_as_video is True, it loads from the video
        annotation file. Otherwise, from the image annotation file.

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        if self.load_as_video:
            data_list, valid_data_indices = self._load_video_data_list()
        else:
            data_list, valid_data_indices = self._load_image_data_list()

        return data_list, valid_data_indices

    def _load_video_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from a video annotation file named as
        ``self.ann_file``.

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_path:
            coco = CocoVID(local_path)
        # The order of returned `cat_ids` will not
        # change with the order of the CLASSES
        self.cat_ids = coco.get_cat_ids(cat_names=self.metainfo['CLASSES'])
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.cat_img_map = copy.deepcopy(coco.cat_img_map)

        data_list = []
        valid_data_indices = []
        data_id = 0
        vid_ids = coco.get_vid_ids()
        for vid_id in vid_ids:
            img_ids = coco.get_img_ids_from_vid(vid_id)
            for img_id in img_ids:
                # load img info
                raw_img_info = coco.load_imgs([img_id])[0]
                raw_img_info['img_id'] = img_id
                raw_img_info['video_length'] = len(img_ids)

                # load ann info
                ann_ids = coco.get_ann_ids(
                    img_ids=[img_id], cat_ids=self.cat_ids)
                raw_ann_info = coco.load_anns(ann_ids)

                if (self.key_img_sampler is not None) and (
                        raw_img_info['frame_id'] %
                        self.key_img_sampler.get('interval', 1) == 0):
                    valid_data_indices.append(data_id)
                # get data_info
                parsed_data_info = self.parse_data_info(
                    dict(raw_img_info=raw_img_info, raw_ann_info=raw_ann_info))
                data_list.append(parsed_data_info)
                data_id += 1

        return data_list, valid_data_indices

    def _load_image_data_list(self) -> Tuple[List[dict], List]:
        """Load annotations from an image annotation file named as
        ``self.ann_file``.

        Returns:
            tuple(list[dict], list): A list of annotation and a list of
            valid data indices.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_path:
            coco = COCO(local_path)
        # The order of returned `cat_ids` will not
        # change with the order of the CLASSES
        self.cat_ids = coco.get_cat_ids(cat_names=self.metainfo['CLASSES'])
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.cat_img_map = copy.deepcopy(coco.cat_img_map)

        data_list = []
        valid_data_indices = []
        data_id = 0
        img_ids = coco.get_img_ids()
        total_ann_ids = []
        for img_id in img_ids:
            # load img info
            raw_img_info = coco.load_imgs([img_id])[0]
            raw_img_info['img_id'] = img_id

            # load ann info
            ann_ids = coco.get_ann_ids(img_ids=[img_id], cat_ids=self.cat_ids)
            raw_ann_info = coco.load_anns(ann_ids)
            total_ann_ids.extend(ann_ids)

            # load images for training
            valid_data_indices.append(data_id)

            # get data_info
            parsed_data_info = self.parse_data_info(
                dict(raw_img_info=raw_img_info, raw_ann_info=raw_ann_info))
            data_list.append(parsed_data_info)
            data_id += 1

        assert len(set(total_ann_ids)) == len(
            total_ann_ids
        ), f"Annotation ids in '{self.ann_file}' are not unique!"

        return data_list, valid_data_indices

    def parse_data_info(self, raw_data_info: dict) -> dict:
        """Parse raw annotation to target format.

        Args:
            raw_data_info (dict): Raw data information load from ``ann_file``

        Returns:
            dict: Parsed annotation.
        """
        img_info = raw_data_info['raw_img_info']
        ann_info = raw_data_info['raw_ann_info']
        data_info = {}

        data_info.update(img_info)
        if self.data_prefix.get('img_path', None) is not None:
            img_path = osp.join(self.data_prefix['img_path'],
                                img_info['file_name'])
        else:
            img_path = img_info['file_name']
        data_info['img_path'] = img_path

        instances = []
        for i, ann in enumerate(ann_info):
            instance = {}

            if ann.get('ignore', False):
                continue
            x1, y1, w, h = ann['bbox']
            inter_w = max(0, min(x1 + w, img_info['width']) - max(x1, 0))
            inter_h = max(0, min(y1 + h, img_info['height']) - max(y1, 0))
            if inter_w * inter_h == 0:
                continue
            if ann['area'] <= 0 or w < 1 or h < 1:
                continue
            if ann['category_id'] not in self.cat_ids:
                continue
            bbox = [x1, y1, x1 + w, y1 + h]

            if ann.get('iscrowd', False):
                instance['ignore_flag'] = 1
            else:
                instance['ignore_flag'] = 0
            instance['bbox'] = bbox
            instance['bbox_label'] = self.cat2label[ann['category_id']]
            if ann.get('segmentation', None):
                instance['mask'] = ann['segmentation']
            if ann.get('instance_id', None):
                instance['instance_id'] = ann['instance_id']
            else:
                # image dataset usually has no `instance_id`.
                # Therefore, we set it to `i`.
                instance['instance_id'] = i
            if len(instance) > 0:
                instances.append(instance)
        data_info['instances'] = instances
        return data_info

    def filter_data(self) -> List[int]:
        """Filter annotations according to filter_cfg.

        Returns:
            list[int]: Filtered results.
        """
        if self.test_mode:
            return self.valid_data_indices

        # obtain images that contain annotation
        ids_with_ann = set(data_info['img_id'] for data_info in self.data_list
                           if len(data_info['instances']) > 0)
        # obtain images that contain annotations of the required categories
        ids_in_cat = set()
        for i, class_id in enumerate(self.cat_ids):
            ids_in_cat |= set(self.cat_img_map[class_id])
        # merge the image id sets of the two conditions and use the merged set
        # to filter out images if self.filter_empty_gt=True
        ids_in_cat &= ids_with_ann

        valid_data_indices = []
        for i, data_info in enumerate(self.data_list):
            img_id = data_info['img_id']
            width = data_info['width']
            height = data_info['height']
            if self.filter_cfg is None:
                if img_id not in ids_in_cat:
                    continue
                if min(width, height) >= 32:
                    valid_data_indices.append(i)
            else:
                if self.filter_cfg.get('filter_empty_gt',
                                       True) and img_id not in ids_in_cat:
                    continue
                if min(width, height) >= self.filter_cfg.get('min_size', 32):
                    valid_data_indices.append(i)

        set_valid_data_indices = set(self.valid_data_indices)
        valid_data_indices = [
            id for id in valid_data_indices if id in set_valid_data_indices
        ]
        return valid_data_indices

    def get_cat_ids(self, idx: int) -> List[int]:
        """Get category ids by index.

        Args:
            idx (int): Index of data.

        Returns:
            List[int]: All categories in the image of specified index.
        """

        instances = self.get_data_info(idx)['instances']
        return [instance['bbox_label'] for instance in instances]

    @force_full_init
    def get_data_info(self, idx: int) -> dict:
        """Get annotation by the index of `self.valid_data_indices` and
        automatically call ``full_init`` if the dataset has not been fully
        initialized.

        Args:
            idx (int): The index of data in `self.valid_data_indices`.

        Returns:
            dict: The idx-th annotation of the dataset.
        """
        ori_idx = self.valid_data_indices[idx]
        data_info = super().get_data_info(ori_idx)
        # Reset the `sample_idx`
        # Some codebase needs `sample_idx` of data information. Here we convert
        # the idx to a positive number and save it in data information.
        if idx >= 0:
            data_info['sample_idx'] = idx
        else:
            data_info['sample_idx'] = len(self) + idx
        return data_info

    @force_full_init
    def _get_ori_data_info(self, ori_idx: int) -> dict:
        """Get annotation by the index of `self.data_list` and automatically
        call ``full_init`` if the dataset has not been fully initialized.

        Args:
            ori_idx (int): The index of data in `self.data_list`.

        Returns:
            dict: The ori_idx-th annotation of the `self.data_list``.
        """
        ori_data_info = super().get_data_info(ori_idx)
        # delete the `sample_idx` key
        ori_data_info.pop('sample_idx')
        return ori_data_info

    @force_full_init
    def __len__(self) -> int:
        """Get the length of filtered dataset and automatically call
        ``full_init`` if the  dataset has not been fully init.

        Returns:
            int: The length of filtered dataset.
        """
        return len(self.valid_data_indices)

    def prepare_data(self, idx: int) -> Any:
        """Get data processed by ``self.pipeline``.

        Args:
            idx (int): The index of ``data_info``.

        Returns:
            Any: Depends on ``self.pipeline``.
        """
        data_info = self.get_data_info(idx)
        if self.ref_img_sampler is not None:
            data_infos = self.ref_img_sampling(idx, data_info,
                                               **self.ref_img_sampler)
            for _data in data_infos:
                if 'video_id' in data_infos[0]:
                    assert data_infos[0]['video_id'] == _data['video_id']
                _data['is_video_data'] = self.load_as_video
            final_data_info = data_infos[0].copy()
            # Collate data_list scatters (list of dict to dict of list)
            for key in final_data_info.keys():
                final_data_info[key] = [_data[key] for _data in data_infos]
        else:
            final_data_info = data_info.copy()
            final_data_info['is_video_data'] = self.load_as_video

        return self.pipeline(final_data_info)

    def ref_img_sampling(self,
                         idx: int,
                         data_info: dict,
                         frame_range: list,
                         stride: int = 1,
                         num_ref_imgs: int = 1,
                         filter_key_img: bool = True,
                         method: str = 'uniform') -> List[dict]:
        """Sampling reference frames in the same video for key frame.

        Args:
            idx (int): The index of `data_info`.
            data_info (dict): The information of key frame.
            frame_range (List(int) | int): The sampling range of reference
                frames in the same video for key frame.
            stride (int): The sampling frame stride when sampling reference
                images. Default: 1.
            num_ref_imgs (int): The number of sampled reference images.
                Default: 1.
            filter_key_img (bool): If False, the key image will be in the
                sampling reference candidates, otherwise, it is exclude.
                Default: True.
            method (str): The sampling method. Options are 'uniform',
                'bilateral_uniform', 'test_with_adaptive_stride',
                'test_with_fix_stride'. 'uniform' denotes reference images are
                randomly sampled from the nearby frames of key frame.
                'bilateral_uniform' denotes reference images are randomly
                sampled from the two sides of the nearby frames of key frame.
                'test_with_adaptive_stride' is only used in testing, and
                denotes the sampling frame stride is equal to (video length /
                the number of reference images). test_with_fix_stride is only
                used in testing with sampling frame stride equalling to
                `stride`. Default: 'uniform'.

        Returns:
            list[dict]: `data_info` and the reference images information.
        """
        assert isinstance(data_info, dict)
        if isinstance(frame_range, int):
            assert frame_range >= 0, 'frame_range can not be a negative value.'
            frame_range = [-frame_range, frame_range]
        elif isinstance(frame_range, list):
            assert len(frame_range) == 2, 'The length must be 2.'
            assert frame_range[0] <= 0 and frame_range[1] >= 0
            for i in frame_range:
                assert isinstance(i, int), 'Each element must be int.'
        else:
            raise TypeError('The type of frame_range must be int or list.')

        if 'test' in method and \
                (frame_range[1] - frame_range[0]) != num_ref_imgs:
            logger = MMLogger.get_current_instance()
            logger.info(
                'Warning:'
                "frame_range[1] - frame_range[0] isn't equal to num_ref_imgs."
                'Set num_ref_imgs to frame_range[1] - frame_range[0].')
            self.ref_img_sampler[
                'num_ref_imgs'] = frame_range[1] - frame_range[0]

        if (not self.load_as_video) or data_info.get('frame_id', -1) < 0 \
                or (frame_range[0] == 0 and frame_range[1] == 0):
            ref_data_infos = []
            for i in range(num_ref_imgs):
                ref_data_infos.append(data_info.copy())
        else:
            frame_id = data_info['frame_id']
            left = max(0, frame_id + frame_range[0])
            right = min(frame_id + frame_range[1],
                        data_info['video_length'] - 1)
            frame_ids = list(range(0, data_info['video_length']))

            ref_frame_ids = []
            if method == 'uniform':
                valid_ids = frame_ids[left:right + 1]
                if filter_key_img and frame_id in valid_ids:
                    valid_ids.remove(frame_id)
                num_samples = min(num_ref_imgs, len(valid_ids))
                ref_frame_ids.extend(random.sample(valid_ids, num_samples))
            elif method == 'bilateral_uniform':
                assert num_ref_imgs % 2 == 0, \
                    'only support load even number of ref_imgs.'
                for mode in ['left', 'right']:
                    if mode == 'left':
                        valid_ids = frame_ids[left:frame_id + 1]
                    else:
                        valid_ids = frame_ids[frame_id:right + 1]
                    if filter_key_img and frame_id in valid_ids:
                        valid_ids.remove(frame_id)
                    num_samples = min(num_ref_imgs // 2, len(valid_ids))
                    sampled_inds = random.sample(valid_ids, num_samples)
                    ref_frame_ids.extend(sampled_inds)
            elif method == 'test_with_adaptive_stride':
                if frame_id == 0:
                    stride = float(len(frame_ids) - 1) / (num_ref_imgs - 1)
                    for i in range(num_ref_imgs):
                        ref_id = round(i * stride)
                        ref_frame_ids.append(frame_ids[ref_id])
            elif method == 'test_with_fix_stride':
                if frame_id == 0:
                    for i in range(frame_range[0], 1):
                        ref_frame_ids.append(frame_ids[0])
                    for i in range(1, frame_range[1] + 1):
                        ref_id = min(round(i * stride), len(frame_ids) - 1)
                        ref_frame_ids.append(frame_ids[ref_id])
                elif frame_id % stride == 0:
                    ref_id = min(
                        round(frame_id + frame_range[1] * stride),
                        len(frame_ids) - 1)
                    ref_frame_ids.append(frame_ids[ref_id])
                data_info['num_left_ref_imgs'] = abs(frame_range[0])
                data_info['frame_stride'] = stride
            else:
                raise NotImplementedError

            ref_data_infos = []
            for ref_frame_id in ref_frame_ids:
                offset = ref_frame_id - frame_id
                ref_data_info = self._get_ori_data_info(
                    self.valid_data_indices[idx] + offset)

                # We need data_info and ref_data_info to have the same keys.
                for key in data_info.keys():
                    if key not in ref_data_info:
                        ref_data_info[key] = data_info[key]

                ref_data_infos.append(ref_data_info)

            ref_data_infos = sorted(
                ref_data_infos, key=lambda i: i['frame_id'])
        return [data_info, *ref_data_infos]



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from os.path import dirname, exists, join

import numpy as np
import torch
from mmdet.utils.util_random import ensure_rng
from mmengine.dataset import pseudo_collate
from mmengine.structures import InstanceData

from mmtrack.structures import TrackDataSample


def random_boxes(num=1, scale=1, rng=None):
    """Simple version of ``kwimage.Boxes.random``

    Returns:
        Tensor: shape (n, 4) in x1, y1, x2, y2 format.

    References:
        https://gitlab.kitware.com/computer-vision/kwimage/blob/master/kwimage/structs/boxes.py#L1390 # noqa: E501

    Example:
        >>> num = 3
        >>> scale = 512
        >>> rng = 0
        >>> boxes = random_boxes(num, scale, rng)
        >>> print(boxes)
        tensor([[280.9925, 278.9802, 308.6148, 366.1769],
                [216.9113, 330.6978, 224.0446, 456.5878],
                [405.3632, 196.3221, 493.3953, 270.7942]])
    """
    rng = ensure_rng(rng)

    tlbr = rng.rand(num, 4).astype(np.float32)

    tl_x = np.minimum(tlbr[:, 0], tlbr[:, 2])
    tl_y = np.minimum(tlbr[:, 1], tlbr[:, 3])
    br_x = np.maximum(tlbr[:, 0], tlbr[:, 2])
    br_y = np.maximum(tlbr[:, 1], tlbr[:, 3])

    tlbr[:, 0] = tl_x * scale
    tlbr[:, 1] = tl_y * scale
    tlbr[:, 2] = br_x * scale
    tlbr[:, 3] = br_y * scale

    boxes = torch.from_numpy(tlbr)
    return boxes


def _get_config_directory():
    """Find the predefined video detector or tracker config directory."""
    try:
        # Assume we are running in the source mmtracking repo
        repo_dpath = dirname(dirname(dirname(__file__)))
    except NameError:
        # For IPython development when this __file__ is not defined
        import mmtrack
        repo_dpath = dirname(dirname(mmtrack.__file__))
    config_dpath = join(repo_dpath, 'configs')
    if not exists(config_dpath):
        raise Exception('Cannot find config path')
    return config_dpath


def _get_config_module(fname):
    """Load a configuration as a python module."""
    from mmengine import Config
    config_dpath = _get_config_directory()
    config_fpath = join(config_dpath, fname)
    config_mod = Config.fromfile(config_fpath)
    return config_mod


def get_model_cfg(fname):
    """Grab configs necessary to create a video detector or tracker.

    These are deep copied to allow for safe modification of parameters without
    influencing other tests.
    """
    config = _get_config_module(fname)
    model = copy.deepcopy(config.model)
    return model


def _rand_bboxes(rng, num_boxes, w, h):
    cx, cy, bw, bh = rng.rand(num_boxes, 4).T

    tl_x = ((cx * w) - (w * bw / 2)).clip(0, w)
    tl_y = ((cy * h) - (h * bh / 2)).clip(0, h)
    br_x = ((cx * w) + (w * bw / 2)).clip(0, w)
    br_y = ((cy * h) + (h * bh / 2)).clip(0, h)

    bboxes = np.vstack([tl_x, tl_y, br_x, br_y]).T
    return bboxes


def _rand_masks(rng, num_boxes, bboxes, img_w, img_h):
    from mmdet.structures.mask import BitmapMasks
    masks = np.zeros((num_boxes, img_h, img_w))
    for i, bbox in enumerate(bboxes):
        bbox = bbox.astype(np.int32)
        mask = (rng.rand(1, bbox[3] - bbox[1], bbox[2] - bbox[0]) >
                0.3).astype(np.int)
        masks[i:i + 1, bbox[1]:bbox[3], bbox[0]:bbox[2]] = mask
    return BitmapMasks(masks, height=img_h, width=img_w)


def demo_mm_inputs(batch_size=1,
                   frame_id=0,
                   num_key_imgs=1,
                   num_ref_imgs=1,
                   image_shapes=[(3, 128, 128)],
                   num_items=None,
                   num_classes=10,
                   ref_prefix='ref',
                   num_template_imgs=None,
                   num_search_imgs=None,
                   with_mask=False,
                   with_semantic=False):
    """Create a superset of inputs needed to run test or train batches.

    Args:
        batch_size (int): batch size. Default to 2.
        frame_id (int): the frame id.
        num_key_imgs (int): the number of key images. This input is used in
            all methods except for training in SOT.
        num_ref_imgs (int): the number of reference images. This input is
            used in all methods except for training in SOT.
        image_shapes (List[tuple], Optional): image shape.
            Default to (3, 128, 128)
        num_items (None | List[int]): specifies the number
            of boxes in each batch item. Default to None.
        num_classes (int): number of different labels a
            box might have. Default to 10.
        ref_prefix (str): the prefix of reference images (or search images
            in SOT).
        with_mask (bool): Whether to return mask annotation.
            Defaults to False.
        with_semantic (bool): whether to return semantic.
            Default to False.
        num_template_imgs (int): the number of template images. This input is
            only used in training in SOT.
        num_search_imgs (int): the number of search images. This input is
            only used in training in SOT.
    """
    # Compatible the names of one image group in SOT. `ref_prefix` means the
    # prefix of search images in SOT.
    assert (num_template_imgs is None) == (num_search_imgs is None)
    if num_template_imgs is not None:
        num_key_imgs, num_ref_imgs = num_template_imgs, num_search_imgs

    rng = np.random.RandomState(0)

    # Make sure the length of image_shapes is equal to ``batch_size``
    if isinstance(image_shapes, list):
        assert len(image_shapes) == batch_size
    else:
        image_shapes = [image_shapes] * batch_size

    # Make sure the length of each element in image_shapes is equal to
    # the number of types of image shapes since key_img and ref_image may have
    # different shapes.
    # After these transforms, as for ``image_shapes``, the
    # length of the outer list is equal to ``batch_size`` and the length of the
    # inner list is equal to the type of image shapes.
    num_img_group = int((num_key_imgs > 0) + (num_ref_imgs > 0))
    if isinstance(image_shapes[0], list):
        assert len(image_shapes[0]) == num_img_group and isinstance(
            image_shapes[0][0], tuple)
    else:
        assert isinstance(image_shapes[0], tuple)
        image_shapes = [[shape] * num_img_group for shape in image_shapes]

    if isinstance(num_items, list):
        assert len(num_items) == batch_size

    packed_inputs = []
    for idx in range(batch_size):
        image_shape_group = image_shapes[idx]
        c, h, w = image_shape_group[0]

        mm_inputs = dict(inputs=dict())
        if num_key_imgs > 0:
            key_img = rng.randint(
                0,
                255,
                size=(num_key_imgs, *image_shape_group[0]),
                dtype=np.uint8)
            mm_inputs['inputs']['img'] = torch.from_numpy(key_img)
        if num_ref_imgs > 0:
            index = int(num_key_imgs > 0)
            ref_img = rng.randint(
                0,
                255,
                size=(num_ref_imgs, *image_shape_group[index]),
                dtype=np.uint8)
            mm_inputs['inputs'][f'{ref_prefix}_img'] = torch.from_numpy(
                ref_img)

        img_meta = {
            'img_id': idx,
            'img_shape': image_shape_group[0][-2:],
            'ori_shape': image_shape_group[0][-2:],
            'filename': '<demo>.png',
            'scale_factor': np.array([1.1, 1.2]),
            'flip': False,
            'flip_direction': None,
            'is_video_data': True,
            'frame_id': frame_id
        }
        if num_ref_imgs > 0:
            search_img_meta = dict()
            for key, value in img_meta.items():
                search_img_meta[f'{ref_prefix}_{key}'] = [
                    value
                ] * num_ref_imgs if num_ref_imgs > 1 else value
            search_shape = image_shape_group[int(num_key_imgs > 0)][-2:]
            search_img_meta[f'{ref_prefix}_img_shape'] = [
                search_shape
            ] * num_ref_imgs if num_ref_imgs > 1 else search_shape
            search_img_meta[f'{ref_prefix}_ori_shape'] = [
                search_shape
            ] * num_ref_imgs if num_ref_imgs > 1 else search_shape
            img_meta.update(search_img_meta)

        data_sample = TrackDataSample()
        data_sample.set_metainfo(img_meta)

        # gt_instances
        gt_instances = InstanceData()
        if num_items is None:
            num_boxes = rng.randint(1, 10)
        else:
            num_boxes = num_items[idx]

        bboxes = _rand_bboxes(rng, num_boxes, w, h)
        labels = rng.randint(0, num_classes, size=num_boxes)
        instances_id = rng.randint(100, num_classes + 100, size=num_boxes)
        gt_instances.bboxes = torch.FloatTensor(bboxes)
        gt_instances.labels = torch.LongTensor(labels)
        gt_instances.instances_id = torch.LongTensor(instances_id)

        if with_mask:
            masks = _rand_masks(rng, num_boxes, bboxes, w, h)
            gt_instances.masks = masks

        data_sample.gt_instances = gt_instances
        # ignore_instances
        ignore_instances = InstanceData()
        bboxes = _rand_bboxes(rng, num_boxes, w, h)
        ignore_instances.bboxes = bboxes
        data_sample.ignored_instances = ignore_instances

        if num_ref_imgs > 0:
            ref_gt_instances = copy.deepcopy(gt_instances)
            setattr(data_sample, f'{ref_prefix}_gt_instances',
                    ref_gt_instances)
            ref_ignored_instances = copy.deepcopy(ignore_instances)
            setattr(data_sample, f'{ref_prefix}_ignored_instances',
                    ref_ignored_instances)

        mm_inputs['data_samples'] = data_sample

        # TODO: gt_ignore

        packed_inputs.append(mm_inputs)
    data = pseudo_collate(packed_inputs)
    return data



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
import tempfile
from argparse import ArgumentParser

import mmcv
import mmengine
import numpy as np


def parse_args():
    parser = ArgumentParser()
    parser.add_argument('video_path', help='video path')
    parser.add_argument('track_results', help='the tracked results')
    parser.add_argument('--gt_bboxes', help='the groundtruth bboxes file')
    parser.add_argument('--output', help='output video file (mp4 format)')
    parser.add_argument(
        '--show',
        action='store_true',
        default=False,
        help='whether to show visualizations.')
    parser.add_argument('--fps', help='FPS of the output video')
    args = parser.parse_args()
    return args


def main(args):

    # load images
    if osp.isdir(args.video_path):
        imgs = sorted(
            filter(lambda x: x.endswith(('.jpg', '.png', '.jpeg')),
                   os.listdir(args.video_path)),
            key=lambda x: int(x.split('.')[0]))
        IN_VIDEO = False
    else:
        imgs = mmcv.VideoReader(args.video_path)
        IN_VIDEO = True

    OUT_VIDEO = False
    # define output
    if args.output is not None:
        if args.output.endswith('.mp4'):
            OUT_VIDEO = True
            out_dir = tempfile.TemporaryDirectory()
            out_path = out_dir.name
            _out = args.output.rsplit(os.sep, 1)
            if len(_out) > 1:
                os.makedirs(_out[0], exist_ok=True)
        else:
            out_path = args.output
            os.makedirs(out_path, exist_ok=True)
    fps = args.fps
    if args.show or OUT_VIDEO:
        if fps is None:
            if IN_VIDEO:
                fps = imgs.fps
            if OUT_VIDEO:
                raise ValueError('Please set the FPS for the output video.')
        else:
            fps = int(fps)

    prog_bar = mmengine.ProgressBar(len(imgs))
    track_bboxes = mmengine.list_from_file(args.track_results)
    if args.gt_bboxes is not None:
        gt_bboxes = mmengine.list_from_file(args.gt_bboxes)
        assert len(track_bboxes) == len(gt_bboxes)

    # test and show/save the images
    for i, img in enumerate(imgs):
        if isinstance(img, str):
            img_path = osp.join(args.video_path, img)
            img = mmcv.imread(img_path)

        if args.output is not None:
            if IN_VIDEO or OUT_VIDEO:
                out_file = osp.join(out_path, f'{i:06d}.jpg')
            else:
                out_file = osp.join(out_path, img_path.rsplit(os.sep, 1)[-1])
        else:
            out_file = None

        draw_bboxes = []
        track_bbox = np.array(list(map(float,
                                       track_bboxes[i].split(','))))[None]
        track_bbox[:, 2] += track_bbox[:, 0]
        track_bbox[:, 3] += track_bbox[:, 1]
        draw_bboxes.append(track_bbox)
        colors = 'green'
        if args.gt_bboxes is not None:
            gt_bbox = np.array(list(map(float, gt_bboxes[i].split(','))))[None]
            gt_bbox[:, 2] += gt_bbox[:, 0]
            gt_bbox[:, 3] += gt_bbox[:, 1]
            draw_bboxes.append(gt_bbox)
            colors = ['green', 'blue']

        mmcv.imshow_bboxes(
            img,
            draw_bboxes,
            show=args.show,
            colors=colors,
            wait_time=int(1000. / fps) if fps else 0,
            out_file=out_file,
            thickness=2)
        prog_bar.update()

    if args.output and OUT_VIDEO:
        print(
            f'\nmaking the output video at {args.output} with a FPS of {fps}')
        mmcv.frames2video(out_path, args.output, fps=fps, fourcc='mp4v')
        out_dir.cleanup()


if __name__ == '__main__':
    args = parse_args()
    main(args)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Tuple

from mmdet.models.roi_heads.roi_extractors import \
    SingleRoIExtractor as _SingleRoIExtractor
from torch import Tensor

from mmtrack.registry import MODELS


@MODELS.register_module()
class SingleRoIExtractor(_SingleRoIExtractor):
    """Extract RoI features from a single level feature map.

    This Class is the same as `SingleRoIExtractor` from
    `mmdet.models.roi_heads.roi_extractors` except for using `**kwargs` to
    accept external arguments.
    """

    def forward(self,
                feats: Tuple[Tensor],
                rois: Tensor,
                roi_scale_factor: float = None,
                **kwargs) -> Tensor:
        """Forward function.
        Args:
            feats (Tuple[Tensor]): The feature maps.
            rois (Tensor): The RoIs.
            roi_scale_factor (float): Scale factor that RoI will be multiplied
                by.

        Returns:
            Tensor: RoI features.
        """
        return super().forward(feats, rois, roi_scale_factor)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

from mmengine.hooks import Hook
from mmengine.model import is_model_wrapper
from torch.nn.modules.batchnorm import BatchNorm2d

from mmtrack.registry import HOOKS


@HOOKS.register_module()
class SiamRPNBackboneUnfreezeHook(Hook):
    """Start to train the backbone of SiamRPN++ from a certrain epoch.

    Args:
        backbone_start_train_epoch (int): Start to train the backbone at
            `backbone_start_train_epoch`-th epoch. Note the epoch in this
            class counts from 0, while the epoch in the log file counts from 1.
        backbone_train_layers (list(str)): List of str denoting the stages
            needed be trained in backbone.
    """

    def __init__(self,
                 backbone_start_train_epoch: int = 10,
                 backbone_train_layers: List = ['layer2', 'layer3', 'layer4']):
        self.backbone_start_train_epoch = backbone_start_train_epoch
        self.backbone_train_layers = backbone_train_layers

    def before_train_epoch(self, runner):
        """If `runner.epoch >= self.backbone_start_train_epoch`, start to train
        the backbone."""
        if runner.epoch >= self.backbone_start_train_epoch:
            for layer in self.backbone_train_layers:
                model = runner.model.module if is_model_wrapper(
                    runner.model) else runner.model
                for param in getattr(model.backbone, layer).parameters():
                    param.requires_grad = True
                for m in getattr(model.backbone, layer).modules():
                    if isinstance(m, BatchNorm2d):
                        m.train()



# Copyright (c) OpenMMLab. All rights reserved.
from .anchor import *  # noqa: F401,F403
from .filter import *  # noqa: F401,F403
from .motion import *  # noqa: F401,F403
from .track import *  # noqa: F401,F403



# Copyright (c) OpenMMLab. All rights reserved.
from numbers import Number
from typing import Dict, List, Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn
from mmdet.structures.mask import BitmapMasks
from mmengine.model import BaseDataPreprocessor

from mmtrack.registry import MODELS
from mmtrack.structures import TrackDataSample
from mmtrack.utils import stack_batch


@MODELS.register_module()
class TrackDataPreprocessor(BaseDataPreprocessor):
    """Image pre-processor for tracking tasks.

    Accepts the data sampled by the dataloader, and preprocesses it into the
    format of the model input. ``TrackDataPreprocessor`` provides the
    tracking data pre-processing as follows:

    - Collate and move data to the target device.
    - Pad inputs to the maximum size of current batch with defined
      ``pad_value``. The padding size can be divisible by a defined
      ``pad_size_divisor``
    - Stack inputs to inputs.
    - Convert inputs from bgr to rgb if the shape of input is (1, 3, H, W).
    - Normalize image with defined std and mean.
    - Do batch augmentations during training.
    - Record the information of ``batch_input_shape`` and ``pad_shape``.

    Args:
        mean (Sequence[Number], optional): The pixel mean of R, G, B channels.
            Defaults to None.
        std (Sequence[Number], optional): The pixel standard deviation of
            R, G, B channels. Defaults to None.
        pad_size_divisor (int): The size of padded image should be
            divisible by ``pad_size_divisor``. Defaults to 1.
        pad_value (Number): The padded pixel value. Defaults to 0.
        pad_mask (bool): Whether to pad instance masks. Defaults to False.
        mask_pad_value (int): The padded pixel value for instance masks.
            Defaults to 0.
        bgr_to_rgb (bool): whether to convert image from BGR to RGB.
            Defaults to False.
        rgb_to_bgr (bool): whether to convert image from RGB to RGB.
            Defaults to False.
        batch_augments (list[dict], optional): Batch-level augmentations
    """

    def __init__(self,
                 mean: Sequence[Number] = None,
                 std: Sequence[Number] = None,
                 pad_size_divisor: int = 1,
                 pad_value: Union[float, int] = 0,
                 pad_mask: bool = False,
                 mask_pad_value: int = 0,
                 bgr_to_rgb: bool = False,
                 rgb_to_bgr: bool = False,
                 batch_augments: Optional[List[dict]] = None):
        super().__init__()
        assert not (bgr_to_rgb and rgb_to_bgr), (
            '`bgr2rgb` and `rgb2bgr` cannot be set to True at the same time')
        assert (mean is None) == (std is None), (
            'mean and std should be both None or tuple')
        if mean is not None:
            assert len(mean) == 3 or len(mean) == 1, (
                'The length of mean should be 1 or 3 to be compatible with '
                f'RGB or gray image, but got {len(mean)}')
            assert len(std) == 3 or len(std) == 1, (  # type: ignore
                'The length of std should be 1 or 3 to be compatible with RGB '  # type: ignore # noqa: E501
                f'or gray image, but got {len(std)}')

            # Enable the normalization in preprocessing.
            self._enable_normalize = True
            self.register_buffer('mean',
                                 torch.tensor(mean).view(1, -1, 1, 1), False)
            self.register_buffer('std',
                                 torch.tensor(std).view(1, -1, 1, 1), False)
        else:
            self._enable_normalize = False

        self.channel_conversion = rgb_to_bgr or bgr_to_rgb
        self.pad_size_divisor = pad_size_divisor
        self.pad_value = pad_value
        self.pad_mask = pad_mask
        self.mask_pad_value = mask_pad_value
        if batch_augments is not None:
            self.batch_augments = nn.ModuleList(
                [MODELS.build(aug) for aug in batch_augments])
        else:
            self.batch_augments = None

    def forward(self, data: dict, training: bool = False) -> Dict:
        """Perform normalization、padding and bgr2rgb conversion based on
        ``TrackDataPreprocessor``.

        Args:
            data (dict): data sampled from dataloader.
            training (bool): Whether to enable training time augmentation.

        Returns:
            Tuple[Dict[str, List[torch.Tensor]], OptSampleList]: Data in the
            same format as the model input.
        """
        batch_pad_shape = self._get_pad_shape(data)
        data = super().forward(data=data, training=training)
        ori_inputs, data_samples = data['inputs'], data['data_samples']

        inputs = dict()
        for imgs_key, imgs in ori_inputs.items():
            # TODO: whether normalize should be after stack_batch
            # imgs is a list contain multiple Tensor of imgs.
            # The shape of imgs[0] is (T, C, H, W).
            channel = imgs[0].size(1)
            if self.channel_conversion and channel == 3:
                imgs = [_img[:, [2, 1, 0], ...] for _img in imgs]
            # change to `float`
            imgs = [_img.float() for _img in imgs]
            if self._enable_normalize:
                imgs = [(_img - self.mean) / self.std for _img in imgs]

            inputs[imgs_key] = stack_batch(imgs, self.pad_size_divisor,
                                           self.pad_value)

        if data_samples is not None:
            # NOTE the batched image size information may be useful, e.g.
            # in DETR, this is needed for the construction of masks, which is
            # then used for the transformer_head.
            for key, imgs in inputs.items():
                img_shape = tuple(imgs.size()[-2:])
                imgs_shape = [img_shape] * imgs.size(1) if imgs.size(
                    1) > 1 else img_shape
                ref_prefix = key[:-3]
                for data_sample, pad_shapes in zip(data_samples,
                                                   batch_pad_shape[key]):
                    data_sample.set_metainfo({
                        f'{ref_prefix}batch_input_shape':
                        imgs_shape,
                        f'{ref_prefix}pad_shape':
                        pad_shapes
                    })
                if self.pad_mask:
                    self.pad_gt_masks(data_samples, ref_prefix)

        if training and self.batch_augments is not None:
            for batch_aug in self.batch_augments:
                # Only yolox need batch_aug, and yolox can only process
                # `img` key. Therefore, only img is processed here.
                # The shape of `img` is (N, T, C, H, W), hence, we use
                # [:, 0] to change the shape to (N, C, H, W).
                assert len(inputs) == 1 and 'img' in inputs
                aug_inputs, data_samples = batch_aug(inputs['img'][:, 0],
                                                     data_samples)
                inputs['img'] = aug_inputs.unsqueeze(1)

        return dict(inputs=inputs, data_samples=data_samples)

    def _get_pad_shape(self, data: dict) -> Dict[str, List]:
        """Get the pad_shape of each image based on data and pad_size_divisor.

        Args:
            data (dict): Data sampled from dataloader.

        Returns:
            Dict[str, List]: The shape of padding.
        """
        batch_pad_shape = dict()
        for imgs_key in data['inputs']:
            pad_shape_list = []
            for imgs in data['inputs'][imgs_key]:
                pad_h = int(
                    np.ceil(imgs.shape[-2] /
                            self.pad_size_divisor)) * self.pad_size_divisor
                pad_w = int(
                    np.ceil(imgs.shape[-1] /
                            self.pad_size_divisor)) * self.pad_size_divisor
                pad_shapes = [
                    (pad_h, pad_w)
                ] * imgs.size(0) if imgs.size(0) > 1 else (pad_h, pad_w)
                pad_shape_list.append(pad_shapes)
            batch_pad_shape[imgs_key] = pad_shape_list
        return batch_pad_shape

    def pad_gt_masks(self, data_samples: Sequence[TrackDataSample],
                     ref_prefix: str) -> None:
        """Pad gt_masks to shape of batch_input_shape."""
        if 'masks' in data_samples[0].get(f'{ref_prefix}gt_instances'):
            for data_samples in data_samples:
                masks = data_samples.get(f'{ref_prefix}gt_instances').masks
                assert isinstance(masks, BitmapMasks)
                batch_input_shape = data_samples.get(
                    f'{ref_prefix}batch_input_shape')
                # handle cases where the number of image > 1
                if isinstance(batch_input_shape, list):
                    batch_input_shape = batch_input_shape[0]
                data_samples.get(
                    f'{ref_prefix}gt_instances').masks = masks.pad(
                        batch_input_shape, pad_val=self.mask_pad_value)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp

import mmengine
import numpy as np

from mmtrack.utils import (plot_norm_precision_curve, plot_precision_curve,
                           plot_success_curve)


def main():
    parser = argparse.ArgumentParser(description='sot plot')
    parser.add_argument(
        'sot_eval_res',
        help='the json/yaml/pickle file path of evaluation results. It can '
        "be a file or path directory. If it's a path directory, all the "
        'files in this directory will be loaded. The final loaded '
        'content must be a collection of name/value pairs. The '
        'name is a tracker name. The value is also a collection of name/value '
        'pairs in the format dict(success=np.ndarray, '
        'norm_precision=np.ndarray, precision=np.ndarray). The metrics have '
        'shape (M, ), where M is the number of values corresponding to '
        'different thresholds.')
    parser.add_argument(
        '--show',
        action='store_true',
        default=False,
        help='whether to show the plotting results')
    parser.add_argument(
        '--plot_save_path',
        default=None,
        type=str,
        help='The saved path of the figure.')
    args = parser.parse_args()

    if osp.isdir(args.sot_eval_res):
        all_eval_results = dict()
        for res_file in os.listdir(args.sot_eval_res):
            if res_file.endswith(('.json', '.yml', '.yaml', '.pkl')):
                eval_res = mmengine.load(osp.join(args.sot_eval_res, res_file))
                all_eval_results.update(eval_res)
    else:
        assert osp.isfile(
            args.sot_eval_res), f'The file {args.sot_eval_res} does not exist'
        all_eval_results = mmengine.load(args.sot_eval_res)
    assert isinstance(all_eval_results, dict)

    tracker_names = []
    all_success = []
    all_norm_precision = []
    all_precision = []
    for tracker_name, scores in all_eval_results.items():
        tracker_names.append(tracker_name)
        if 'success' in scores:
            all_success.append(scores['success'])
        if 'precision' in scores:
            all_precision.append(scores['precision'])
        if 'norm_precision' in scores:
            all_norm_precision.append(scores['norm_precision'])

    if len(all_success) > 0:
        all_success = np.stack(all_success)
        plot_success_curve(
            all_success,
            tracker_names=tracker_names,
            plot_save_path=args.plot_save_path,
            show=args.show)
    if len(all_precision) > 0:
        all_precision = np.stack(all_precision)
        plot_precision_curve(
            all_precision,
            tracker_names=tracker_names,
            plot_save_path=args.plot_save_path,
            show=args.show)
    if len(all_norm_precision) > 0:
        all_norm_precision = np.stack(all_norm_precision)
        plot_norm_precision_curve(
            all_norm_precision,
            tracker_names=tracker_names,
            plot_save_path=args.plot_save_path,
            show=args.show)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import os
import time
from typing import List

import numpy as np

from mmtrack.registry import DATASETS
from .base_sot_dataset import BaseSOTDataset


@DATASETS.register_module()
class VOTDataset(BaseSOTDataset):
    """VOT dataset of single object tracking. The dataset is only used to test.

    Args:
        dataset_type (str, optional): The type of VOT challenge. The
            optional values are in ['vot2018', 'vot2018_lt',
            'vot2019', 'vot2019_lt']
    """

    def __init__(self, dataset_type: str = 'vot2018', *args, **kwargs):
        """Initialization of SOT dataset class."""
        assert dataset_type in [
            'vot2018', 'vot2018_lt', 'vot2019', 'vot2019_lt'
        ], 'We only support VOT-[2018~2019] chanllenges'
        self.dataset_type = dataset_type
        super().__init__(*args, **kwargs)

    def load_data_list(self) -> List[dict]:
        """Load dataset information.

        Returns:
            list[dict]: The length of the list is the number of videos. The
                inner dict is in the following format:
                    {
                        'video_path': the video path
                        'ann_path': the annotation path
                        'start_frame_id': the starting frame number contained
                            in the image name
                        'end_frame_id': the ending frame number contained in
                            the image name
                        'framename_template': the template of image name
                    }
        """
        print('Loading VOT dataset...')
        start_time = time.time()
        data_infos = []
        data_infos_str = self._loadtxt(
            self.ann_file, return_ndarray=False).split('\n')
        # the first line of annotation file is a dataset comment.
        for line in data_infos_str[1:]:
            # compatible with different OS.
            line = line.strip().replace('/', os.sep).split(',')
            data_info = dict(
                video_path=line[0],
                ann_path=line[1],
                start_frame_id=int(line[2]),
                end_frame_id=int(line[3]),
                framename_template='%08d.jpg')
            data_infos.append(data_info)
        print(f'VOT dataset loaded! ({time.time()-start_time:.2f} s)')
        return data_infos

    def get_ann_infos_from_video(self, video_ind: int) -> np.ndarray:
        """Get bboxes annotation about the instance in a video.

        Args:
            video_ind (int): video index

        Returns:
            np.ndarray: in [N, 8] shape. The N is the bbox number and the bbox
                is in (x1, y1, x2, y2, x3, y3, x4, y4) format.
        """
        bboxes = self.get_bboxes_from_video(video_ind)
        if bboxes.shape[1] == 4:
            x1, y1 = bboxes[:, 0], bboxes[:, 1],
            x2, y2 = bboxes[:, 0] + bboxes[:, 2], bboxes[:, 1],
            x3, y3 = bboxes[:, 0] + bboxes[:, 2], bboxes[:, 1] + bboxes[:, 3]
            x4, y4 = bboxes[:, 0], bboxes[:, 1] + bboxes[:, 3],
            bboxes = np.stack((x1, y1, x2, y2, x3, y3, x4, y4), axis=-1)

        visible_info = self.get_visibility_from_video(video_ind)
        # bboxes in VOT datasets are all valid
        bboxes_isvalid = np.array([True] * len(bboxes), dtype=np.bool_)
        ann_infos = dict(
            bboxes=bboxes, bboxes_isvalid=bboxes_isvalid, **visible_info)
        return ann_infos



# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from typing import Optional, Sequence

from mmdet.datasets.api_wrappers import COCO
from mmdet.evaluation import CocoMetric
from mmdet.structures.mask import encode_mask_results
from mmengine.dist import broadcast_object_list, is_main_process
from mmengine.fileio import FileClient

from mmtrack.registry import METRICS
from .base_video_metrics import collect_tracking_results


@METRICS.register_module()
class CocoVideoMetric(CocoMetric):
    """COCO evaluation metric.

    Evaluate AR, AP, and mAP for detection tasks including proposal/box
    detection and instance segmentation. Please refer to
    https://cocodataset.org/#detection-eval for more details.
    """

    def __init__(self, ann_file: Optional[str] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        # if ann_file is not specified,
        # initialize coco api with the converted dataset
        if ann_file:
            file_client = FileClient.infer_client(uri=ann_file)
            with file_client.get_local_path(ann_file) as local_path:
                self._coco_api = COCO(local_path)
        else:
            self._coco_api = None

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.results``, which will be used to
        compute the metrics when all batches have been processed.

        Note that we only modify ``pred['pred_instances']`` in ``CocoMetric``
        to ``pred['pred_det_instances']`` here.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of data samples that
                contain annotations and predictions.
        """
        for data_sample in data_samples:
            result = dict()
            pred = data_sample['pred_det_instances']
            result['img_id'] = data_sample['img_id']
            result['bboxes'] = pred['bboxes'].cpu().numpy()
            result['scores'] = pred['scores'].cpu().numpy()
            result['labels'] = pred['labels'].cpu().numpy()
            # encode mask to RLE
            if 'masks' in pred:
                result['masks'] = encode_mask_results(
                    pred['masks'].detach().cpu().numpy())
            # some detectors use different scores for bbox and mask
            if 'mask_scores' in pred:
                result['mask_scores'] = pred['mask_scores'].cpu().numpy()

            # parse gt
            gt = dict()
            gt['width'] = data_sample['ori_shape'][1]
            gt['height'] = data_sample['ori_shape'][0]
            gt['img_id'] = data_sample['img_id']
            if self._coco_api is None:
                assert 'instances' in data_sample, \
                    'ground truth is required for evaluation when ' \
                    '`ann_file` is not provided'
                gt['anns'] = data_sample['instances']
            # add converted result to the results list
            self.results.append((gt, result))

    def evaluate(self, size: int) -> dict:
        """Evaluate the model performance of the whole dataset after processing
        all batches.

        Args:
            size (int): Length of the entire validation dataset.

        Returns:
            dict: Evaluation metrics dict on the val dataset. The keys are the
            names of the metrics, and the values are corresponding results.
        """
        if len(self.results) == 0:
            warnings.warn(
                f'{self.__class__.__name__} got empty `self.results`. Please '
                'ensure that the processed results are properly added into '
                '`self.results` in `process` method.')

        results = collect_tracking_results(self.results, self.collect_device)

        if is_main_process():
            _metrics = self.compute_metrics(results)  # type: ignore
            # Add prefix to metric names
            if self.prefix:
                _metrics = {
                    '/'.join((self.prefix, k)): v
                    for k, v in _metrics.items()
                }
            metrics = [_metrics]
        else:
            metrics = [None]  # type: ignore

        broadcast_object_list(metrics)

        # reset the results list
        self.results.clear()
        return metrics[0]



# Copyright (c) OpenMMLab. All rights reserved.
from .entire_video_batch_sampler import EntireVideoBatchSampler
from .quota_sampler import QuotaSampler
from .video_sampler import VideoSampler

__all__ = ['VideoSampler', 'QuotaSampler', 'EntireVideoBatchSampler']



_base_ = ['./selsa-troialign_faster-rcnn_r50-dc5_8xb1-7e_imagenetvid.py']
model = dict(
    detector=dict(
        backbone=dict(
            depth=101,
            init_cfg=dict(
                type='Pretrained', checkpoint='torchvision://resnet101'))))



# dataset settings
dataset_type = 'MOTChallengeDataset'
data_root = 'data/MOT17/'

# data pipeline
train_pipeline = [
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=True),
            dict(
                type='mmdet.RandomResize',
                scale=(1088, 1088),
                ratio_range=(0.8, 1.2),
                keep_ratio=True,
                clip_object_border=False),
            dict(type='mmdet.PhotoMetricDistortion')
        ]),
    dict(
        type='TransformBroadcaster',
        share_random_params=False,
        transforms=[
            dict(
                type='mmdet.RandomCrop',
                crop_size=(1088, 1088),
                bbox_clip_border=False)
        ]),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='mmdet.RandomFlip', prob=0.5),
        ]),
    dict(type='PackTrackInputs', ref_prefix='ref', num_key_frames=1)
]

test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadTrackAnnotations', with_instance_id=True),
    dict(type='mmdet.Resize', scale=(1088, 1088), keep_ratio=True),
    dict(type='PackTrackInputs', pack_single_img=True)
]

# dataloader
train_dataloader = dict(
    batch_size=2,
    num_workers=2,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    batch_sampler=dict(type='mmdet.AspectRatioBatchSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        visibility_thr=-1,
        ann_file='annotations/half-train_cocoformat.json',
        data_prefix=dict(img_path='train'),
        metainfo=dict(CLASSES=('pedestrian', )),
        ref_img_sampler=dict(
            num_ref_imgs=1,
            frame_range=10,
            filter_key_img=True,
            method='uniform'),
        pipeline=train_pipeline))
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='VideoSampler'),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        ann_file='annotations/half-val_cocoformat.json',
        data_prefix=dict(img_path='train'),
        ref_img_sampler=None,
        load_as_video=True,
        test_mode=True,
        pipeline=test_pipeline))
test_dataloader = val_dataloader

# evaluator
val_evaluator = dict(
    type='MOTChallengeMetrics', metric=['HOTA', 'CLEAR', 'Identity'])
test_evaluator = val_evaluator



_base_ = ['./stark-st1_r50_8xb16-500e_got10k-lasot-trackingnet-coco_base.py']

# model setting
model = dict(
    type='Stark',
    head=dict(
        type='StarkHead',
        cls_head=dict(
            type='ScoreHead',
            input_dim=256,
            hidden_dim=256,
            output_dim=1,
            num_layers=3,
            use_bn=False),
        frozen_modules=['transformer', 'bbox_head', 'query_embedding'],
        loss_cls=dict(type='mmdet.CrossEntropyLoss', use_sigmoid=True)),
    frozen_modules=['backbone', 'neck'])

data_root = {{_base_.data_root}}
# the only difference of ``train_pipeline`` compared with that in stark_st1 is
# ``train_cls_head=True`` in ``TridentSampling``.
train_pipeline = [
    dict(
        type='TridentSampling',
        num_search_frames=1,
        num_template_frames=2,
        max_frame_range=[200],
        cls_pos_prob=0.5,
        train_cls_head=True),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[
            dict(type='LoadImageFromFile'),
            dict(type='LoadTrackAnnotations', with_instance_id=False),
            dict(type='GrayAug', prob=0.05),
            dict(type='mmdet.RandomFlip', prob=0.5, direction='horizontal')
        ]),
    dict(
        type='SeqBboxJitter',
        center_jitter_factor=[0, 0, 4.5],
        scale_jitter_factor=[0, 0, 0.5],
        crop_size_factor=[2, 2, 5]),
    dict(
        type='SeqCropLikeStark',
        crop_size_factor=[2, 2, 5],
        output_size=[128, 128, 320]),
    dict(
        type='TransformBroadcaster',
        share_random_params=True,
        transforms=[dict(type='BrightnessAug', jitter_range=0.2)]),
    dict(type='CheckPadMaskValidity', stride=16),
    dict(type='PackTrackInputs', ref_prefix='search', num_template_frames=2)
]

# dataset settings
train_dataloader = dict(
    dataset=dict(datasets=[
        dict(
            type='GOT10kDataset',
            data_root=data_root,
            ann_file='GOT10k/annotations/got10k_train_vot_infos.txt',
            data_prefix=dict(img_path='GOT10k'),
            pipeline=train_pipeline,
            test_mode=False),
        dict(
            type='LaSOTDataset',
            data_root=data_root,
            ann_file='LaSOT_full/annotations/lasot_train_infos.txt',
            data_prefix=dict(img_path='LaSOT_full/LaSOTBenchmark'),
            pipeline=train_pipeline,
            test_mode=False),
        dict(
            type='TrackingNetDataset',
            data_root=data_root,
            ann_file='TrackingNet/annotations/trackingnet_train_infos.txt',
            data_prefix=dict(img_path='TrackingNet'),
            pipeline=train_pipeline,
            test_mode=False),
        dict(
            type='SOTCocoDataset',
            data_root=data_root,
            ann_file='coco/annotations/instances_train2017.json',
            data_prefix=dict(img_path='coco/train2017'),
            pipeline=train_pipeline,
            test_mode=False)
    ]))

# runner loop
train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=50, val_begin=50, val_interval=1)

# learning policy
param_scheduler = dict(type='MultiStepLR', milestones=[40], gamma=0.1)

# checkpoint saving
default_hooks = dict(checkpoint=dict(type='CheckpointHook', interval=10))



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmcv.ops import PrRoIPool
from mmengine.model import BaseModule
from torch import Tensor, nn

from mmtrack.registry import MODELS


@MODELS.register_module()
class FilterInitializer(BaseModule):
    """Initializes a target classification filter.

    Args:
        filter_size (int, optional):  Size of the filter. Defaults to 4.
        feature_dim (int, optional):  Input feature dimentionality.
             Defaults to 512.
        feature_stride (int, optional):  Input feature stride. Defaults to 16.
    """

    def __init__(self,
                 filter_size: int = 4,
                 feature_dim: int = 512,
                 feature_stride: int = 16):
        super().__init__()
        self.filter_conv = nn.Conv2d(
            feature_dim, feature_dim, kernel_size=3, padding=1)
        self.filter_pool = PrRoIPool(filter_size, 1 / feature_stride)

    def init_weights(self):
        """Initialize the parameters of this module."""
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                m.weight.data.zero_()
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, feat: Tensor, bboxes: Tensor) -> Tensor:
        """Runs the initializer module. Note that [] denotes an optional
        dimension.

        Args:
            feat (Tensor):  Input feature maps with shape
                (images_in_sequence, [sequences], feat_dim, H, W).
            bboxes (Tensor):  Target bounding boxes with
                (images_in_sequence, [sequences], 4) shape in [x1, y1, x2, y2]
                format.

        Returns:
            filter_weights (Tensor):  The output filter with shape
                (images_in_sequence, c, filter_h, filter_w).
        """

        num_images = feat.shape[0]

        feat = self.filter_conv(feat.reshape(-1, *feat.shape[-3:]))

        # Add batch_index to rois
        batch_index = torch.arange(
            bboxes.shape[0], dtype=torch.float32).reshape(-1,
                                                          1).to(bboxes.device)
        roi = torch.cat((batch_index, bboxes), dim=1)
        filter_weights = self.filter_pool(feat, roi)

        # If multiple input images, compute the initial filter
        # as the average filter.
        if num_images > 1:
            filter_weights = torch.mean(
                filter_weights.reshape(num_images, -1,
                                       *filter_weights.shape[-3:]),
                dim=0)

        return filter_weights
